"""
********************************************************************************
* Copyright (c) 2026 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

from functools import lru_cache

import jax
from jax import make_jaxpr
from jax.extend.core import Jaxpr, Literal, ClosedJaxpr
from jax.tree_util import tree_flatten, tree_unflatten

from qrisp.jasp import (
    eval_jaxpr,
    flatten_environments,
    cond_to_cl_control,
    extract_invalues,
    flatten_environments,
    insert_outvalues,
)
from qrisp.jasp.primitives import (
    AbstractQuantumState,
    QuantumPrimitive,
    ParityOperation,
)
from qrisp.jasp.interpreter_tools.interpreters import ProcessedMeasurement
from qrisp.jasp.primitives import AbstractQuantumState, QuantumPrimitive
from qrisp.jasp.jasp_expression import collect_environments, invert_jaspr


class Jaspr(ClosedJaxpr):
    """
    The ``Jaspr`` class enables an efficient representations of a wide variety
    of (hybrid) algorithms. For many applications, the representation is agnostic
    to the scale of the problem, implying function calls with 10 or 10000 qubits
    can be represented by the same object. The actual unfolding to a circuit-level
    description is outsourced to
    `established, classical compilation infrastructure <https://mlir.llvm.org/>`_,
    implying state-of-the-art compilation speed can be reached.

    As a subtype of ``jax.extend.core.ClosedJaxpr``, Jasprs are embedded into the well matured
    `Jax ecosystem <https://github.com/n2cholas/awesome-jax>`_,
    which facilitates the compilation of classical `real-time computation <https://arxiv.org/abs/2206.12950>`_
    using some of the most advanced libraries in the world such as
    `CUDA <https://jax.readthedocs.io/en/latest/Custom_Operation_for_GPUs.html>`_.
    Especially `machine learning <https://ai.google.dev/gemma/docs/jax_inference>`_
    and other scientific computations tasks are particularly well supported.

    To get a better understanding of the syntax and semantics of Jaxpr (and with
    that also Jaspr) please check `this link <https://jax.readthedocs.io/en/latest/jaxpr.html>`__.

    Similar to Jaxpr, Jaspr objects represent (hybrid) quantum
    algorithms in the form of a `functional programming language <https://en.wikipedia.org/wiki/Functional_programming>`_
    in `SSA-form <https://en.wikipedia.org/wiki/Static_single-assignment_form>`_.

    It is possible to compile Jaspr objects into QIR, which is facilitated by the
    `Catalyst framework <https://docs.pennylane.ai/projects/catalyst/en/stable/index.html>`__
    (check :meth:`qrisp.jasp.jaspr.to_qir` for more details).

    Qrisp scripts can be turned into Jaspr objects by
    calling the ``make_jaspr`` function, which has similar semantics as
    `jax.make_jaxpr <https://jax.readthedocs.io/en/latest/_autosummary/jax.make_jaxpr.html>`_.

    ::

        from qrisp import *
        from qrisp.jasp import make_jaspr

        def test_fun(i):

            qv = QuantumFloat(i, -1)
            x(qv[0])
            cx(qv[0], qv[i-1])
            meas_res = measure(qv)
            meas_res += 1
            return meas_res


        jaspr = make_jaspr(test_fun)(4)
        print(jaspr)

    This will give you the following output:

    .. code-block::

        { lambda ; a:QuantumState b:i32[]. let
            c:QuantumState d:QubitArray = create_qubits a b
            e:Qubit = get_qubit d 0
            f:QuantumState = x c e
            g:i32[] = sub b 1
            h:Qubit = get_qubit d g
            i:QuantumState = cx f e h
            j:QuantumState k:i32[] = measure i d
            l:f32[] = convert_element_type[new_dtype=float64 weak_type=True] k
            m:f32[] = mul l 0.5
            n:f32[] = add m 1.0
          in (j, n) }


    A defining feature of the Jaspr class is that the first input and the
    first output are always of QuantumState type. Therefore, Jaspr objects always
    represent some (hybrid) quantum operation.

    Qrisp comes with a built-in Jaspr interpreter. For that you simply have to
    call the object like a function:


    >>> print(jaspr(2))
    2.5
    >>> print(jaspr(4))
    5.5
    """

    __slots__ = (
        "permeability",
        "isqfree",
        "hashvalue",
        "ctrl_jaspr",
        "inv_jaspr",
        "envs_flattened",
    )

    def __init__(
        self,
        *args,
        permeability=None,
        isqfree=None,
        ctrl_jaspr=None,
        inv_jaspr=None,
        **kwargs,
    ):

        if len(args) == 2:
            if not isinstance(args[0], Jaxpr) or not isinstance(args[1], list):
                raise Exception(
                    f"Tried to call the Jaspr constructor with two arguments and signature {type(args[0]), type(args[1])} (allowed is (Jaxpr, list))"
                )
            kwargs["jaxpr"] = args[0]
            kwargs["consts"] = args[1]

        elif len(args) == 1:
            if not isinstance(args[0], ClosedJaxpr):
                raise Exception(
                    f"Tried to call the Jaspr constructor with one argument and signature {type(args[0])} (allowed is (ClosedJaxpr))"
                )
            kwargs["jaxpr"] = args[0].jaxpr
            kwargs["consts"] = args[0].consts

        if "jaxpr" in kwargs:

            ClosedJaxpr.__init__(self, kwargs["jaxpr"], kwargs["consts"])
        else:

            if "consts" in kwargs:
                consts = kwargs["consts"]
                del kwargs["consts"]
            else:
                if len(kwargs["constvars"]):
                    raise Exception(
                        "Tried to create Jaspr with constvars but no constants"
                    )
                consts = []

            ClosedJaxpr.__init__(self, jaxpr=Jaxpr(**kwargs), consts=consts)

        self.hashvalue = id(self)
        self.permeability = {}
        if permeability is None:
            permeability = {}
        for var in self.constvars + self.invars + self.outvars:
            if isinstance(var, Literal):
                continue
            self.permeability[var] = permeability.get(var, None)

        self.isqfree = isqfree
        self.ctrl_jaspr = ctrl_jaspr
        self.inv_jaspr = inv_jaspr
        self.envs_flattened = False

        if not isinstance(self.invars[-1].aval, AbstractQuantumState):
            raise Exception(
                f"Tried to create a Jaspr from data that doesn't have a QuantumState the last argument (got {type(self.invars[-1].aval)} instead)"
            )

        if not isinstance(self.outvars[-1].aval, AbstractQuantumState):
            raise Exception(
                f"Tried to create a Jaspr from data that doesn't have a QuantumState the last entry of return type (got {type(self.outvars[-1].aval)} instead)"
            )

    @property
    def constvars(self):
        return self.jaxpr.constvars

    @property
    def eqns(self):
        return self.jaxpr.eqns

    @property
    def invars(self):
        return self.jaxpr.invars

    @property
    def outvars(self):
        return self.jaxpr.outvars

    @property
    def debug_info(self):
        return self.jaxpr.debug_info

    def __hash__(self):
        return self.hashvalue

    def __eq__(self, other):
        if not isinstance(other, Jaxpr):
            return False

        return id(self) == id(other)

    def copy(self):

        if self.ctrl_jaspr is None:
            ctrl_jaspr = None
        else:
            ctrl_jaspr = self.ctrl_jaspr.copy()

        res = Jaspr(
            permeability=self.permeability,
            isqfree=self.isqfree,
            ctrl_jaspr=ctrl_jaspr,
            constvars=list(self.constvars),
            invars=list(self.invars),
            outvars=list(self.outvars),
            eqns=list(self.eqns),
            effects=self.effects,
            debug_info=self.debug_info,
        )

        res.envs_flattened = self.envs_flattened

        return res

    def inverse(self):
        """
        Returns the inverse Jaspr (if applicable). For Jaspr that contain realtime
        computations or measurements, the inverse does not exist.

        Returns
        -------
        Jaspr
            The daggered Jaspr.

        Examples
        --------

        We create a simple script and inspect the daggered version:

        ::

            from qrisp import *
            from qrisp.jasp import make_jaspr

            def example_function(i):

                qv = QuantumVariable(i)
                cx(qv[0], qv[1])
                t(qv[1])
                return qv

            jaspr = make_jaspr(example_function)(2)

            print(jaspr.inverse())
            # Yields
            # { lambda ; a:QuantumState b:i32[]. let
            #     c:QuantumState d:QubitArray = create_qubits a b
            #     e:Qubit = get_qubit d 0
            #     f:Qubit = get_qubit d 1
            #     g:QuantumState = t_dg c f
            #     h:QuantumState = cx g e f
            #   in (h, d) }
        """
        return invert_jaspr(self)

    def control(self, num_ctrl, ctrl_state=-1):
        """
        Returns the controlled version of the Jaspr. The control qubits are added
        to the signature of the Jaspr as the arguments after the QuantumState.

        Parameters
        ----------
        num_ctrl : int
            The amount of controls to be added.
        ctrl_state : int of str, optional
            The control state on which to activate. The default is -1.

        Returns
        -------
        Jaspr
            The controlled Jaspr.

        Examples
        --------

        We create a simple script and inspect the controlled version:

        ::

            from qrisp import *
            from qrisp.jasp import make_jaspr

            def example_function(i):

                qv = QuantumVariable(i)
                cx(qv[0], qv[1])
                t(qv[1])
                return qv

            jaspr = make_jaspr(example_function)(2)

            print(jaspr.control(2))
            # Yields
            # { lambda ; a:QuantumState b:Qubit c:Qubit d:i32[]. let
            #     e:QuantumState f:QubitArray = create_qubits a 1
            #     g:Qubit = get_qubit f 0
            #     h:QuantumState = ccx e b c g
            #     i:QuantumState j:QubitArray = create_qubits h d
            #     k:Qubit = get_qubit j 0
            #     l:Qubit = get_qubit j 1
            #     m:QuantumState = ccx i g k l
            #     n:QuantumState = ct m g l
            #     o:QuantumState = ccx n b c g
            #   in (o, j) }

        We see that the control qubits are part of the function signature
        (``a`` and ``b``)

        """
        if self.ctrl_jaspr is not None and num_ctrl == 1 and ctrl_state == -1:
            return self.ctrl_jaspr

        from qrisp.jasp import ControlledJaspr

        if isinstance(ctrl_state, int):
            if ctrl_state < 0:
                ctrl_state += 2**num_ctrl

            ctrl_state = bin(ctrl_state)[2:].zfill(num_ctrl)
        else:
            ctrl_state = str(ctrl_state)

        return ControlledJaspr.from_cache(self, ctrl_state)

    def to_qc(self, *args):
        """
        Converts the Jaspr into a :ref:`QuantumCircuit` if applicable. Circuit
        conversion of algorithms involving realtime computations is not possible.

        Any computations that perform classical postprocessing of measurements
        can not be reflected within the QuantumCircuit object itself and will
        generate an object of type ``ProcessedMeasurement``. These objects hold
        no further information and are simply used as placeholders to emulate
        the computation.

        Parameters
        ----------
        *args : tuple
            The arguments to call the Jaspr with.

        Returns
        -------
        return_values : tuple
            The return values of the Jaspr. QuantumVariable return types are
            returned as lists of Qubits.
        :ref:`QuantumCircuit`
            The resulting QuantumCircuit.

        Examples
        --------

        We create a simple script and inspect the QuantumCircuit:

        ::

            from qrisp import *
            from qrisp.jasp import make_jaspr

            def example_function(i):

                qv = QuantumVariable(i)
                cx(qv[0], qv[1])
                t(qv[1])
                return qv

            jaspr = make_jaspr(example_function)(2)

            qb_list, qc = jaspr.to_qc(2)
            print(qc)
            # Yields
            # qb_0: ──■───────
            #       ┌─┴─┐┌───┐
            # qb_1: ┤ X ├┤ T ├
            #       └───┘└───┘

        To demonstrate the behavior under measurement post-processing, we build
        a similar script:

        ::

            from qrisp import ProcessedMeasurement

            def example_function(i):

                qf = QuantumFloat(i)
                cx(qf[0], qf[1])
                t(qf[1])

                meas_res = measure(qf)
                # Perform classical post processing
                meas_res *= 2
                return meas_res

            jaspr = make_jaspr(example_function)(2)

            meas_res, qc = jaspr.to_qc(2)
            print(isinstance(meas_res, ProcessedMeasurement))
            # True



        """
        from qrisp.jasp.interpreter_tools.interpreters import jaspr_to_qc

        return jaspr_to_qc(self, *args)

    def extract_post_processing(self, *args):
        """
        Extracts the post-processing logic from this Jaspr and returns a function
        that performs the post-processing on measurement results.

        This method is useful for separating the quantum circuit from the classical
        post-processing of measurement results. The quantum circuit can be executed
        on a NISQ-style backend to obtain measurement results, and then the post-processing
        function can be applied to those results to obtain the final output.

        .. note::

            It is not possible to extract QuantumCircuits from Jaspr objects
            involving real-time computation, but it is possible to extract a post
            processing function.

        Parameters
        ----------
        *args : tuple
            The static argument values that were used for circuit extraction.
            These will be bound into the post-processing function as Literals.

        Returns
        -------
        callable
            A function that takes measurement results and returns the post-processed results.
            Accepts either a string of '0' and '1' characters or a JAX array of booleans
            with shape (n,). String inputs are automatically converted to boolean arrays.

        Examples
        --------

        We create a Jaspr that performs post-processing on measurement results:

        ::

            from qrisp import *
            import jax.numpy as jnp

            @make_jaspr
            def example_function(i):
                qv = QuantumFloat(5)
                # First measurement
                meas_1 = measure(qv[i])
                h(qv[1])
                # Second measurement
                meas_2 = measure(qv[1])
                # Classical post-processing
                return meas_1 + 2, meas_2

            jaspr = example_function(1)

            # Extract the quantum circuit
            a, b, qc = jaspr.to_qc(1)

            # Extract the post-processing function with the SAME arguments
            post_proc = jaspr.extract_post_processing(1)

            # Execute qc on a backend to get measurement results
            results = qc.run()

            # Apply post-processing to each result
            for bitstring, count in results.items():
                processed = post_proc(bitstring)
                print(f"{bitstring} -> {processed}")

            # Yields:
            # 00 -> (Array(2, dtype=int64), Array(False, dtype=bool))
            # 01 -> (Array(2, dtype=int64), Array(True, dtype=bool))

            # Can also use with array input (useful for JAX jitting):
            import jax.numpy as jnp
            meas_array = jnp.array([False, True])
            processed = post_proc(meas_array)

        Note that the static arguments (in this case `1`) must be the same as those
        used for circuit extraction, since they affect the structure of both the
        quantum circuit and the post-processing logic.
        """
        from qrisp.jasp.interpreter_tools.interpreters import extract_post_processing

        return extract_post_processing(self, *args)

    def eval(self, *args, eqn_evaluator=lambda x, y: True):
        return eval_jaxpr(self, eqn_evaluator=eqn_evaluator)(*args)

    def flatten_environments(self):
        """
        Flattens all environments by applying the corresponding compilation
        routines such that no more ``q_env`` primitives are left.

        Returns
        -------
        Jaspr
            The Jaspr with flattened environments.

        Examples
        --------

        We create a Jaspr containing an :ref:`InversionEnvironment` and flatten:

        ::

                def test_function(i):
                        qv = QuantumVariable(i)

                        with invert():
                                t(qv[0])
                                cx(qv[0], qv[1])

                        return qv

                jaspr = make_jaspr(test_function)(2)
                print(jaspr)

        ::

                { lambda ; a:QuantumState b:i32[]. let
                        c:QuantumState d:QubitArray = create_qubits a b
                        e:QuantumState = q_env[
                        jaspr={ lambda ; f:QuantumState d:QubitArray. let
                  g:Qubit = get_qubit d 0
                  h:QuantumState = t f g
                  i:Qubit = get_qubit d 1
                  j:QuantumState = cx h g i
                in (j,) }
                        type=InversionEnvironment
                        ] c d
                  in (e, d) }

        You can see how the body of the :ref:`InversionEnvironment` is __collected__
        into another Jaspr. This reflects the fact that at their core,
        :ref:`QuantumEnvironment <QuantumEnvironment>` describe `higher-order
        quantum functions <https://en.wikipedia.org/wiki/Higher-order_function>`_
        (ie. functions that operate on functions). In order to apply the
        transformations induced by the QuantumEnvironment, we can call
        ``jaspr.flatten_environments``:

        >>> print(jaspr.flatten_environments)
        { lambda ; a:QuantumState b:i32[]. let
            c:QuantumState d:QubitArray = create_qubits a b
            e:Qubit = get_qubit d 0
            f:Qubit = get_qubit d 1
            g:QuantumState = cx c e f
            h:QuantumState = t_dg g e
          in (h, d) }

        We see that as expected, the order of the ``cx`` and the ``t`` gate has been switched and the ``t`` gate has been turned into a ``t_dg``.

        """
        res = flatten_environments(self)
        if self.ctrl_jaspr is not None:
            res.ctrl_jaspr = self.ctrl_jaspr.flatten_environments()
        if self.inv_jaspr is not None:
            if not self.inv_jaspr.envs_flattened:
                res.inv_jaspr = self.inv_jaspr.flatten_environments()
            else:
                res.inv_jaspr = self.inv_jaspr
            res.inv_jaspr.inv_jaspr = res
        return res

    def __call__(self, *args):
        from qrisp.jasp.evaluation_tools.jaspification import simulate_jaspr

        return simulate_jaspr(self, *args)

        if len(self.outvars) == 1:
            return None

        from qrisp.simulator import BufferedQuantumState

        args = [BufferedQuantumState()] + list(tree_flatten(args)[0])

        from qrisp.jasp import eval_jaxpr, extract_invalues, insert_outvalues

        flattened_jaspr = self

        def eqn_evaluator(eqn, context_dic):
            if eqn.primitive.name == "jit":

                if eqn.params["name"] == "expectation_value_eval_function":
                    from qrisp.jasp.program_control import sampling_evaluator

                    sampling_evaluator("ev")(
                        eqn, context_dic, eqn_evaluator=eqn_evaluator
                    )
                    return

                if eqn.params["name"] == "sampling_eval_function":
                    from qrisp.jasp.program_control import sampling_evaluator

                    sampling_evaluator("array")(
                        eqn, context_dic, eqn_evaluator=eqn_evaluator
                    )
                    return

                invalues = extract_invalues(eqn, context_dic)
                outvalues = eval_jaxpr(
                    eqn.params["jaxpr"], eqn_evaluator=eqn_evaluator
                )(*invalues)
                if not isinstance(outvalues, (list, tuple)):
                    outvalues = [outvalues]
                insert_outvalues(eqn, context_dic, outvalues)
            elif eqn.primitive.name == "jasp.create_quantum_kernel":
                insert_outvalues(eqn, context_dic, BufferedQuantumState())
            else:
                return True

        res = eval_jaxpr(flattened_jaspr, eqn_evaluator=eqn_evaluator)(
            *(args + self.consts)
        )

        if len(self.outvars) == 2:
            return res[1]
        else:
            return res[1:]

    def inline(self, *args):

        from qrisp.jasp import TracingQuantumSession

        qs = TracingQuantumSession.get_instance()
        abs_qst = qs.abs_qst

        ammended_args = list(args) + [abs_qst]
        res = eval_jaxpr(self)(*ammended_args)

        if isinstance(res, tuple):
            new_abs_qst = res[-1]
            res = res[:-1]
        else:
            new_abs_qst = res
            res = None
        qs.abs_qst = new_abs_qst
        return res

    def count_ops(self, *args, meas_behavior):
        from qrisp.jasp.evaluation_tools import profile_jaspr

        return profile_jaspr(self, "count_ops", meas_behavior)(*args)

    def depth(self, *args, meas_behavior, max_qubits=1024):
        from qrisp.jasp.evaluation_tools import profile_jaspr

        return profile_jaspr(self, "depth", meas_behavior, max_qubits=max_qubits)(*args)

    def num_qubits(self, *args, meas_behavior, max_allocations=1000):
        from qrisp.jasp.evaluation_tools import profile_jaspr

        return profile_jaspr(
            self, "num_qubits", meas_behavior, max_allocations=max_allocations
        )(*args)

    def embedd(self, *args, name=None, inline=False):
        from qrisp.jasp import TracingQuantumSession, get_last_equation

        qs = TracingQuantumSession.get_instance()
        abs_qst = qs.abs_qst

        ammended_args = list(args) + [abs_qst]
        if not inline:
            res = jax.jit(eval_jaxpr(self))(*ammended_args)

            eqn = get_last_equation()

            eqn.params["jaxpr"] = self
            if name is not None:
                eqn.params["name"] = name
        else:
            res = eval_jaxpr(self)(*ammended_args)

        if isinstance(res, tuple):
            new_abs_qst = res[-1]
            res = res[:-1]
        else:
            new_abs_qst = res
            res = None
        qs.abs_qst = new_abs_qst
        return res

    def qjit(self, *args, function_name="jaspr_function", device=None):
        """
        Leverages the Catalyst pipeline to compile a QIR representation of
        this function and executes that function using the Catalyst QIR runtime.

        Parameters
        ----------
        *args : iterable
            The arguments to call the function with.
        device : object
            The `PennyLane device <https://docs.pennylane.ai/projects/catalyst/en/stable/dev/devices.html>`_ to execute the function.
            The default device is `"lightning.qubit" <https://docs.pennylane.ai/projects/lightning/en/stable/lightning_qubit/device.html>`_,
            a fast state-vector qubit simulator.

        Returns
        -------
            The values returned by the compiled, executed function.

        """
        flattened_jaspr = self

        from qrisp.jasp.evaluation_tools.catalyst_interface import (
            jaspr_to_catalyst_qjit,
        )

        qjit_obj = jaspr_to_catalyst_qjit(
            flattened_jaspr, function_name=function_name, device=device
        )
        res = qjit_obj.compiled_function(*args)
        if not isinstance(res, (tuple, list)):
            return res
        elif len(res) == 1:
            return res[0]
        else:
            return res

    @classmethod
    @lru_cache(maxsize=int(1e5))
    def from_cache(cls, closed_jaxpr):
        res = Jaspr(jaxpr=closed_jaxpr.jaxpr, consts=closed_jaxpr.consts)
        remove_redundant_allocations(res)
        return res

    def update_eqns(self, eqns):
        return Jaspr(
            constvars=list(self.constvars),
            invars=list(self.invars),
            outvars=list(self.outvars),
            eqns=list(eqns),
            consts=list(self.consts),
            debug_info=self.debug_info,
        )

    def to_qir(self):
        """
        Compiles the Jaspr to QIR using the `Catalyst framework <https://docs.pennylane.ai/projects/catalyst/en/stable/index.html>`__.

        Parameters
        ----------
        None

        Returns
        -------
        str
            The QIR string.

        Examples
        --------

        We create a simple script and inspect the QIR string:

        ::

            from qrisp import *
            from qrisp.jasp import make_jaspr

            def example_function(i):

                qv = QuantumFloat(i)
                cx(qv[0], qv[1])
                t(qv[1])
                meas_res = measure(qv)
                meas_res += 1
                return meas_res

            jaspr = make_jaspr(example_function)(2)
            print(jaspr.to_qir())

        Yields:

        .. code-block:: none

            ; ModuleID = 'LLVMDialectModule'
            source_filename = "LLVMDialectModule"

            @"{'shots': 0, 'mcmc': False, 'num_burnin': 0, 'kernel_name': None}" = internal constant [66 x i8] c"{'shots': 0, 'mcmc': False, 'num_burnin': 0, 'kernel_name': None}\00"
            @lightning.qubit = internal constant [16 x i8] c"lightning.qubit\00"
            @"/home/positr0nium/miniconda3/envs/qrisp/lib/python3.10/site-packages/catalyst/utils/../lib/librtd_lightning.so" = internal constant [111 x i8] c"/home/positr0nium/miniconda3/envs/qrisp/lib/python3.10/site-packages/catalyst/utils/../lib/librtd_lightning.so\00"

            declare void @__catalyst__rt__finalize() local_unnamed_addr

            declare void @__catalyst__rt__initialize() local_unnamed_addr

            declare ptr @__catalyst__qis__Measure(ptr, i32) local_unnamed_addr

            declare void @__catalyst__qis__T(ptr, ptr) local_unnamed_addr

            declare void @__catalyst__qis__CNOT(ptr, ptr, ptr) local_unnamed_addr

            declare ptr @__catalyst__rt__array_get_element_ptr_1d(ptr, i64) local_unnamed_addr

            declare ptr @__catalyst__rt__qubit_allocate_array(i64) local_unnamed_addr

            declare void @__catalyst__rt__device_init(ptr, ptr, ptr) local_unnamed_addr

            declare void @_mlir_memref_to_llvm_free(ptr) local_unnamed_addr

            declare ptr @_mlir_memref_to_llvm_alloc(i64) local_unnamed_addr

            define { ptr, ptr, i64 } @jit_jaspr_function(ptr nocapture readnone %0, ptr nocapture readonly %1, i64 %2) local_unnamed_addr {
              tail call void @__catalyst__rt__device_init(ptr nonnull @"/home/positr0nium/miniconda3/envs/qrisp/lib/python3.10/site-packages/catalyst/utils/../lib/librtd_lightning.so", ptr nonnull @lightning.qubit, ptr nonnull @"{'shots': 0, 'mcmc': False, 'num_burnin': 0, 'kernel_name': None}")
              %4 = tail call ptr @__catalyst__rt__qubit_allocate_array(i64 20)
              %5 = tail call ptr @__catalyst__rt__array_get_element_ptr_1d(ptr %4, i64 0)
              %6 = load ptr, ptr %5, align 8
              %7 = tail call ptr @__catalyst__rt__array_get_element_ptr_1d(ptr %4, i64 1)
              %8 = load ptr, ptr %7, align 8
              tail call void @__catalyst__qis__CNOT(ptr %6, ptr %8, ptr null)
              %9 = tail call ptr @__catalyst__rt__array_get_element_ptr_1d(ptr %4, i64 1)
              %10 = load ptr, ptr %9, align 8
              tail call void @__catalyst__qis__T(ptr %10, ptr null)
              %11 = tail call ptr @_mlir_memref_to_llvm_alloc(i64 65)
              %12 = ptrtoint ptr %11 to i64
              %13 = add i64 %12, 63
              %14 = and i64 %13, -64
              %15 = inttoptr i64 %14 to ptr
              %16 = tail call ptr @_mlir_memref_to_llvm_alloc(i64 65)
              %17 = ptrtoint ptr %16 to i64
              %18 = add i64 %17, 63
              %19 = and i64 %18, -64
              %20 = inttoptr i64 %19 to ptr
              %21 = tail call ptr @_mlir_memref_to_llvm_alloc(i64 72)
              %22 = ptrtoint ptr %21 to i64
              %23 = add i64 %22, 63
              %24 = and i64 %23, -64
              %25 = inttoptr i64 %24 to ptr
              %26 = tail call ptr @_mlir_memref_to_llvm_alloc(i64 72)
              %27 = ptrtoint ptr %26 to i64
              %28 = add i64 %27, 63
              %29 = and i64 %28, -64
              %30 = inttoptr i64 %29 to ptr
              %31 = tail call ptr @_mlir_memref_to_llvm_alloc(i64 72)
              %32 = ptrtoint ptr %31 to i64
              %33 = add i64 %32, 63
              %34 = and i64 %33, -64
              %35 = inttoptr i64 %34 to ptr
              %36 = tail call ptr @_mlir_memref_to_llvm_alloc(i64 72)
              %37 = ptrtoint ptr %36 to i64
              %38 = add i64 %37, 63
              %39 = and i64 %38, -64
              %40 = inttoptr i64 %39 to ptr
              %41 = tail call ptr @_mlir_memref_to_llvm_alloc(i64 8)
              store i64 0, ptr %41, align 1
              %42 = tail call ptr @_mlir_memref_to_llvm_alloc(i64 8)
              store i64 0, ptr %42, align 1
              %43 = load i64, ptr %1, align 4
              %44 = icmp slt i64 %43, 1
              store i1 %44, ptr %15, align 64
              %45 = tail call ptr @_mlir_memref_to_llvm_alloc(i64 8)
              %46 = load i64, ptr %42, align 1
              store i64 %46, ptr %45, align 1
              tail call void @_mlir_memref_to_llvm_free(ptr nonnull %42)
              %47 = tail call ptr @_mlir_memref_to_llvm_alloc(i64 8)
              %48 = load i64, ptr %41, align 1
              store i64 %48, ptr %47, align 1
              tail call void @_mlir_memref_to_llvm_free(ptr nonnull %41)
              br i1 %44, label %.lr.ph, label %._crit_edge

            .lr.ph:                                           ; preds = %3, %.lr.ph
              %49 = phi ptr [ %87, %.lr.ph ], [ %47, %3 ]
              %50 = phi ptr [ %85, %.lr.ph ], [ %45, %3 ]
              %51 = load i64, ptr %50, align 4
              %52 = tail call ptr @__catalyst__rt__array_get_element_ptr_1d(ptr %4, i64 %51)
              %53 = load ptr, ptr %52, align 8
              %54 = tail call ptr @__catalyst__qis__Measure(ptr %53, i32 -1)
              %55 = load i1, ptr %54, align 1
              store i1 %55, ptr %20, align 64
              %56 = load i64, ptr %50, align 4
              store i64 %56, ptr %25, align 64
              %57 = shl i64 2, %56
              %58 = icmp ult i64 %56, 64
              %59 = select i1 %58, i64 %57, i64 0
              store i64 %59, ptr %30, align 64
              %60 = load i1, ptr %20, align 64
              %61 = zext i1 %60 to i64
              store i64 %61, ptr %35, align 64
              %62 = load i64, ptr %30, align 64
              %63 = select i1 %60, i64 %62, i64 0
              store i64 %63, ptr %40, align 64
              %64 = tail call ptr @_mlir_memref_to_llvm_alloc(i64 72)
              %65 = ptrtoint ptr %64 to i64
              %66 = add i64 %65, 63
              %67 = and i64 %66, -64
              %68 = inttoptr i64 %67 to ptr
              %69 = load i64, ptr %49, align 4
              %70 = load i64, ptr %40, align 64
              %71 = add i64 %70, %69
              store i64 %71, ptr %68, align 64
              tail call void @_mlir_memref_to_llvm_free(ptr nonnull %49)
              %72 = tail call ptr @_mlir_memref_to_llvm_alloc(i64 72)
              %73 = ptrtoint ptr %72 to i64
              %74 = add i64 %73, 63
              %75 = and i64 %74, -64
              %76 = inttoptr i64 %75 to ptr
              %77 = load i64, ptr %50, align 4
              %78 = add i64 %77, 1
              store i64 %78, ptr %76, align 64
              tail call void @_mlir_memref_to_llvm_free(ptr nonnull %50)
              %79 = tail call ptr @_mlir_memref_to_llvm_alloc(i64 8)
              %80 = load i64, ptr %68, align 64
              store i64 %80, ptr %79, align 1
              tail call void @_mlir_memref_to_llvm_free(ptr %64)
              %81 = tail call ptr @_mlir_memref_to_llvm_alloc(i64 8)
              %82 = load i64, ptr %76, align 64
              store i64 %82, ptr %81, align 1
              tail call void @_mlir_memref_to_llvm_free(ptr %72)
              %.pre = load i64, ptr %81, align 4
              %83 = load i64, ptr %1, align 4
              %84 = icmp sge i64 %.pre, %83
              store i1 %84, ptr %15, align 64
              %85 = tail call ptr @_mlir_memref_to_llvm_alloc(i64 8)
              %86 = load i64, ptr %81, align 1
              store i64 %86, ptr %85, align 1
              tail call void @_mlir_memref_to_llvm_free(ptr nonnull %81)
              %87 = tail call ptr @_mlir_memref_to_llvm_alloc(i64 8)
              %88 = load i64, ptr %79, align 1
              store i64 %88, ptr %87, align 1
              tail call void @_mlir_memref_to_llvm_free(ptr nonnull %79)
              br i1 %84, label %.lr.ph, label %._crit_edge

            ._crit_edge:                                      ; preds = %.lr.ph, %3
              %.lcssa20 = phi ptr [ %45, %3 ], [ %85, %.lr.ph ]
              %.lcssa = phi ptr [ %47, %3 ], [ %87, %.lr.ph ]
              tail call void @_mlir_memref_to_llvm_free(ptr nonnull %.lcssa20)
              tail call void @_mlir_memref_to_llvm_free(ptr %36)
              tail call void @_mlir_memref_to_llvm_free(ptr %31)
              tail call void @_mlir_memref_to_llvm_free(ptr %26)
              tail call void @_mlir_memref_to_llvm_free(ptr %21)
              tail call void @_mlir_memref_to_llvm_free(ptr %16)
              tail call void @_mlir_memref_to_llvm_free(ptr %11)
              %89 = tail call ptr @_mlir_memref_to_llvm_alloc(i64 68)
              %90 = ptrtoint ptr %89 to i64
              %91 = add i64 %90, 63
              %92 = and i64 %91, -64
              %93 = inttoptr i64 %92 to ptr
              %94 = load i64, ptr %.lcssa, align 4
              %95 = trunc i64 %94 to i32
              store i32 %95, ptr %93, align 64
              tail call void @_mlir_memref_to_llvm_free(ptr nonnull %.lcssa)
              %96 = tail call ptr @_mlir_memref_to_llvm_alloc(i64 68)
              %97 = ptrtoint ptr %96 to i64
              %98 = add i64 %97, 63
              %99 = and i64 %98, -64
              %100 = inttoptr i64 %99 to ptr
              %101 = load i32, ptr %93, align 64
              %102 = add i32 %101, 1
              store i32 %102, ptr %100, align 64
              tail call void @_mlir_memref_to_llvm_free(ptr %89)
              %103 = icmp eq ptr %96, inttoptr (i64 3735928559 to ptr)
              br i1 %103, label %104, label %107

            104:                                              ; preds = %._crit_edge
              %105 = tail call ptr @_mlir_memref_to_llvm_alloc(i64 4)
              %106 = load i32, ptr %100, align 64
              store i32 %106, ptr %105, align 1
              br label %107

            107:                                              ; preds = %104, %._crit_edge
              %.pn16 = phi ptr [ %105, %104 ], [ %96, %._crit_edge ]
              %.pn14 = phi ptr [ %105, %104 ], [ %100, %._crit_edge ]
              %.pn13 = insertvalue { ptr, ptr, i64 } undef, ptr %.pn16, 0
              %.pn = insertvalue { ptr, ptr, i64 } %.pn13, ptr %.pn14, 1
              %108 = insertvalue { ptr, ptr, i64 } %.pn, i64 0, 2
              ret { ptr, ptr, i64 } %108
            }

            define void @_catalyst_pyface_jit_jaspr_function(ptr nocapture writeonly %0, ptr nocapture readonly %1) local_unnamed_addr {
              %.unpack = load ptr, ptr %1, align 8
              %.elt1.i = getelementptr inbounds { ptr, ptr, i64 }, ptr %.unpack, i64 0, i32 1
              %.unpack2.i = load ptr, ptr %.elt1.i, align 8
              %3 = tail call { ptr, ptr, i64 } @jit_jaspr_function(ptr poison, ptr %.unpack2.i, i64 poison)
              %.elt.i = extractvalue { ptr, ptr, i64 } %3, 0
              store ptr %.elt.i, ptr %0, align 8
              %.repack5.i = getelementptr inbounds { ptr, ptr, i64 }, ptr %0, i64 0, i32 1
              %.elt6.i = extractvalue { ptr, ptr, i64 } %3, 1
              store ptr %.elt6.i, ptr %.repack5.i, align 8
              %.repack7.i = getelementptr inbounds { ptr, ptr, i64 }, ptr %0, i64 0, i32 2
              %.elt8.i = extractvalue { ptr, ptr, i64 } %3, 2
              store i64 %.elt8.i, ptr %.repack7.i, align 8
              ret void
            }

            define void @_catalyst_ciface_jit_jaspr_function(ptr nocapture writeonly %0, ptr nocapture readonly %1) local_unnamed_addr {
              %.elt1 = getelementptr inbounds { ptr, ptr, i64 }, ptr %1, i64 0, i32 1
              %.unpack2 = load ptr, ptr %.elt1, align 8
              %3 = tail call { ptr, ptr, i64 } @jit_jaspr_function(ptr poison, ptr %.unpack2, i64 poison)
              %.elt = extractvalue { ptr, ptr, i64 } %3, 0
              store ptr %.elt, ptr %0, align 8
              %.repack5 = getelementptr inbounds { ptr, ptr, i64 }, ptr %0, i64 0, i32 1
              %.elt6 = extractvalue { ptr, ptr, i64 } %3, 1
              store ptr %.elt6, ptr %.repack5, align 8
              %.repack7 = getelementptr inbounds { ptr, ptr, i64 }, ptr %0, i64 0, i32 2
              %.elt8 = extractvalue { ptr, ptr, i64 } %3, 2
              store i64 %.elt8, ptr %.repack7, align 8
              ret void
            }

            define void @setup() local_unnamed_addr {
              tail call void @__catalyst__rt__initialize()
              ret void
            }

            define void @teardown() local_unnamed_addr {
              tail call void @__catalyst__rt__finalize()
              ret void
            }

            !llvm.module.flags = !{!0}

            !0 = !{i32 2, !"Debug Info Version", i32 3}

        """
        from qrisp.jasp.evaluation_tools.catalyst_interface import jaspr_to_qir

        return jaspr_to_qir(self.flatten_environments())

    def to_mlir(self):
        """
        Compiles the Jaspr to an xDSL module using the Jasp Dialect.
        Requires the xDSL package to be installed (``pip install xdsl``).


        .. note::

            An xDSL module can be visualized via:

            ::

                print(xdsl_module)

            and serialized to a string using:

            ::

                from xdsl.printer import Printer
                Printer().print_op(xdsl_module)


        Returns
        -------
        xdsl.dialects.builtin.ModuleOp
            An xDSL module representing the quantum computation.

        Examples
        --------

        We create a simple script and inspect the MLIR string:

        ::

            from qrisp import *
            from qrisp.jasp import make_jaspr

            def example_function(i):

                qv = QuantumVariable(i)
                cx(qv[0], qv[1])
                t(qv[1])
                meas_res = measure(qv)
                meas_res += 1
                return meas_res

            jaspr = make_jaspr(example_function)(2)
            print(jaspr.to_mlir())

        .. code-block:: none

            builtin.module @jasp_module {
              func.func public @main(%arg0 : tensor<i64>, %arg1 : !jasp.QuantumState) -> (tensor<i64>, !jasp.QuantumState) {
                %0, %1 = "jasp.create_qubits"(%arg0, %arg1) : (tensor<i64>, !jasp.QuantumState) -> (!jasp.QubitArray, !jasp.QuantumState)
                %2 = "stablehlo.constant"() <{value = dense<0> : tensor<i64>}> : () -> tensor<i64>
                %3 = "jasp.get_qubit"(%0, %2) : (!jasp.QubitArray, tensor<i64>) -> !jasp.Qubit
                %4 = "stablehlo.constant"() <{value = dense<1> : tensor<i64>}> : () -> tensor<i64>
                %5 = "jasp.get_qubit"(%0, %4) : (!jasp.QubitArray, tensor<i64>) -> !jasp.Qubit
                %6 = "jasp.quantum_gate"(%3, %5, %1) {gate_type = "cx"} : (!jasp.Qubit, !jasp.Qubit, !jasp.QuantumState) -> !jasp.QuantumState
                %7 = "jasp.quantum_gate"(%5, %6) {gate_type = "t"} : (!jasp.Qubit, !jasp.QuantumState) -> !jasp.QuantumState
                %8, %9 = "jasp.measure"(%0, %7) : (!jasp.QubitArray, !jasp.QuantumState) -> (tensor<i64>, !jasp.QuantumState)
                %10 = "stablehlo.add"(%8, %4) : (tensor<i64>, tensor<i64>) -> tensor<i64>
                %11 = "jasp.reset"(%0, %9) : (!jasp.QubitArray, !jasp.QuantumState) -> !jasp.QuantumState
                %12 = "jasp.delete_qubits"(%0, %11) : (!jasp.QubitArray, !jasp.QuantumState) -> !jasp.QuantumState
                func.return %10, %12 : tensor<i64>, !jasp.QuantumState
              }
            }

        """
        from qrisp.jasp.mlir import jaspr_to_mlir

        return jaspr_to_mlir(self)

    def to_catalyst_mlir(self):
        """
        Compiles the Jaspr to MLIR using the `Catalyst dialect <https://docs.pennylane.ai/projects/catalyst/en/stable/index.html>`__.

        Parameters
        ----------
        None

        Returns
        -------
        str
            The MLIR string.

        Examples
        --------

        We create a simple script and inspect the MLIR string:

        ::

            from qrisp import *
            from qrisp.jasp import make_jaspr

            def example_function(i):

                qv = QuantumFloat(i)
                cx(qv[0], qv[1])
                t(qv[1])
                meas_res = measure(qv)
                meas_res += 1
                return meas_res

            jaspr = make_jaspr(example_function)(2)
            print(jaspr.to_catalyst_mlir())

        .. code-block:: none

                        module @jaspr_function {
              func.func public @jit_jaspr_function(%arg0: tensor<i64>) -> tensor<i32> attributes {llvm.emit_c_interface} {
                %0 = stablehlo.constant dense<1> : tensor<i32>
                %1 = stablehlo.constant dense<2> : tensor<i64>
                %2 = stablehlo.constant dense<1> : tensor<i64>
                %3 = stablehlo.constant dense<0> : tensor<i64>
                quantum.device["/home/positr0nium/miniconda3/envs/qrisp/lib/python3.10/site-packages/catalyst/utils/../lib/librtd_lightning.so", "lightning.qubit", "{'shots': 0, 'mcmc': False, 'num_burnin': 0, 'kernel_name': None}"]
                %4 = quantum.alloc( 20) : !quantum.reg
                %5 = quantum.extract %4[ 0] : !quantum.reg -> !quantum.bit
                %6 = quantum.extract %4[ 1] : !quantum.reg -> !quantum.bit
                %out_qubits:2 = quantum.custom "CNOT"() %5, %6 : !quantum.bit, !quantum.bit
                %7 = quantum.insert %4[ 0], %out_qubits#0 : !quantum.reg, !quantum.bit
                %8 = quantum.insert %7[ 1], %out_qubits#1 : !quantum.reg, !quantum.bit
                %9 = quantum.extract %8[ 1] : !quantum.reg -> !quantum.bit
                %out_qubits_0 = quantum.custom "T"() %9 : !quantum.bit
                %10 = quantum.insert %8[ 1], %out_qubits_0 : !quantum.reg, !quantum.bit
                %11 = stablehlo.add %3, %arg0 : tensor<i64>
                %12:3 = scf.while (%arg1 = %3, %arg2 = %3, %arg3 = %10) : (tensor<i64>, tensor<i64>, !quantum.reg) -> (tensor<i64>, tensor<i64>, !quantum.reg) {
                  %16 = stablehlo.compare  GE, %arg1, %11,  SIGNED : (tensor<i64>, tensor<i64>) -> tensor<i1>
                  %extracted = tensor.extract %16[] : tensor<i1>
                  scf.condition(%extracted) %arg1, %arg2, %arg3 : tensor<i64>, tensor<i64>, !quantum.reg
                } do {
                ^bb0(%arg1: tensor<i64>, %arg2: tensor<i64>, %arg3: !quantum.reg):
                  %extracted = tensor.extract %arg1[] : tensor<i64>
                  %16 = quantum.extract %arg3[%extracted] : !quantum.reg -> !quantum.bit
                  %mres, %out_qubit = quantum.measure %16 : i1, !quantum.bit
                  %from_elements = tensor.from_elements %mres : tensor<i1>
                  %extracted_1 = tensor.extract %arg1[] : tensor<i64>
                  %17 = quantum.insert %arg3[%extracted_1], %out_qubit : !quantum.reg, !quantum.bit
                  %18 = stablehlo.subtract %arg1, %3 : tensor<i64>
                  %19 = stablehlo.shift_left %1, %18 : tensor<i64>
                  %20 = stablehlo.convert %from_elements : (tensor<i1>) -> tensor<i64>
                  %21 = stablehlo.multiply %19, %20 : tensor<i64>
                  %22 = stablehlo.add %arg2, %21 : tensor<i64>
                  %23 = stablehlo.add %arg1, %2 : tensor<i64>
                  scf.yield %23, %22, %17 : tensor<i64>, tensor<i64>, !quantum.reg
                }
                %13 = stablehlo.convert %12#1 : (tensor<i64>) -> tensor<i32>
                %14 = stablehlo.multiply %13, %0 : tensor<i32>
                %15 = stablehlo.add %14, %0 : tensor<i32>
                return %15 : tensor<i32>
              }
              func.func @setup() {
                quantum.init
                return
              }
              func.func @teardown() {
                quantum.finalize
                return
              }
            }

        """
        from qrisp.jasp.evaluation_tools.catalyst_interface import jaspr_to_mlir

        return jaspr_to_mlir(self.flatten_environments())

    def to_qasm(self, *args):
        """
        Compiles the Jaspr into an OpenQASM 2 string. Real-time control is possible
        as long as no computations on the measurement results are performed.

        Parameters
        ----------
        *args : list
            The arguments to call the :ref:`QuantumCircuit` evaluation with.

        Returns
        -------
        str
            The OpenQASM 2 string.

        Examples
        --------

        We create a simple script and inspect the QASM 2 string:

        ::

            from qrisp import *
            from qrisp.jasp import make_jaspr

            def main(i):

                qv = QuantumVariable(i)
                cx(qv[0], qv[1])
                t(qv[1])
                return qv

            jaspr = make_jaspr(main)(2)

            qasm_str = jaspr.to_qasm(2)
            print(qasm_str)
            # Yields

            # OPENQASM 2.0;
            # include "qelib1.inc";
            # qreg qb_59[1];
            # qreg qb_60[1];
            # cx qb_59[0],qb_60[0];
            # t qb_60[0];


        It is also possible to compile simple real-time control features:

        ::

            def main(phi):

                qf = QuantumFloat(5)
                h(qf)
                bl = measure(qf[0])

                with control(bl):
                    rz(phi, qf[1])
                    x(qf[1])

                return

            jaspr = make_jaspr(main)(0.5)
            print(jaspr.to_qasm(0.5))

        This gives:

        ::

            OPENQASM 2.0;
            include "qelib1.inc";
            qreg qb_59[1];
            qreg qb_60[1];
            qreg qb_61[1];
            qreg qb_62[1];
            qreg qb_63[1];
            creg cb_0[1];
            h qb_59[0];
            h qb_60[0];
            h qb_61[0];
            reset qb_61[0];
            h qb_62[0];
            reset qb_62[0];
            h qb_63[0];
            reset qb_63[0];
            measure qb_59[0] -> cb_0[0];
            reset qb_59[0];
            if(cb_0==1) rz(0.5) qb_60[0];
            if(cb_0==1) x qb_60[0];
            reset qb_60[0];

        """
        res = self.to_qc(*args)
        if len(self.outvars) == 1:
            res = [res]
        qrisp_qc = res[-1]
        return qrisp_qc.qasm()

    def to_catalyst_jaxpr(self):
        """
        Compiles the jaspr to the corresponding `Catalyst jaxpr <https://docs.pennylane.ai/projects/catalyst/en/stable/index.html>`__.

        Parameters
        ----------
        *args : tuple
            The arguments to call the jaspr with.

        Returns
        -------
        jax.core.Jaxpr
            The Jaxpr using Catalyst primitives.

        Examples
        --------

        We create a simple script and inspect the Catalyst Jaxpr:

        ::

            from qrisp import *
            from qrisp.jasp import make_jaspr

            def example_function(i):

                qv = QuantumFloat(i)
                cx(qv[0], qv[1])
                t(qv[1])
                meas_res = measure(qv)
                meas_res += 1
                return meas_res

            jaspr = make_jaspr(example_function)(2)

            print(jaspr.to_catalyst_jaxpr())
            # Yields
            # { lambda ; a:AbstractQreg() b:i64[] c:i32[]. let
            #   d:i64[] = convert_element_type[new_dtype=int64 weak_type=True] c
            #   e:i64[] = add b d
            #   f:i64[] = add b 0
            #   g:i64[] = add b 1
            #   h:AbstractQbit() = qextract a f
            #   i:AbstractQbit() = qextract a g
            #   j:AbstractQbit() k:AbstractQbit() = qinst[op=CNOT qubits_len=2] h i
            #   l:AbstractQreg() = qinsert a f j
            #   m:AbstractQreg() = qinsert l g k
            #   n:AbstractQbit() = qextract m g
            #   o:AbstractQbit() = qinst[op=T qubits_len=1] n
            #   p:AbstractQreg() = qinsert m g o
            #   q:i64[] = convert_element_type[new_dtype=int64 weak_type=True] c
            #   r:i64[] = add b q
            #   _:i64[] s:i64[] t:AbstractQreg() _:i64[] _:i64[] = while_loop[
            #     body_jaxpr={ lambda ; u:i64[] v:i64[] w:AbstractQreg() x:i64[] y:i64[]. let
            #         z:AbstractQbit() = qextract w u
            #         ba:bool[] bb:AbstractQbit() = qmeasure z
            #         bc:AbstractQreg() = qinsert w u bb
            #         bd:i64[] = sub u x
            #         be:i64[] = shift_left 2 bd
            #         bf:i64[] = convert_element_type[new_dtype=int64 weak_type=True] ba
            #         bg:i64[] = mul be bf
            #         bh:i64[] = add v bg
            #         bi:i64[] = add u 1
            #       in (bi, bh, bc, x, y) }
            #     body_nconsts=0
            #     cond_jaxpr={ lambda ; bj:i64[] bk:i64[] bl:AbstractQreg() bm:i64[] bn:i64[]. let
            #         bo:bool[] = ge bj bn
            #       in (bo,) }
            #     cond_nconsts=0
            #     nimplicit=0
            #     preserve_dimensions=True
            #   ] b 0 p b r
            #   bp:i32[] = convert_element_type[new_dtype=int64 weak_type=False] s
            #   bq:i32[] = mul bp 1
            #   br:i32[] = add bq 1
            # in (t, e, br) }

        """
        from qrisp.jasp.evaluation_tools.catalyst_interface import (
            jaspr_to_catalyst_jaxpr,
        )

        return jaspr_to_catalyst_jaxpr(self.flatten_environments())


def make_jaxpr_mod(fun, static_argnums=(), return_shape=False, abstracted_axes=None):
    """
    Creates a function that produces the jaxpr of a traced function.

    This is a modified version of JAX's ``make_jaxpr`` that supports
    ``return_shape=True`` even when the function returns custom abstract
    types that don't have ``shape``/``dtype`` attributes (such as
    ``AbstractQuantumState``).

    The interface is identical to ``jax.make_jaxpr``.

    Parameters
    ----------
    fun : Callable
        The function whose jaxpr is to be computed.
    static_argnums : int or Sequence[int], optional
        Indices of arguments that should be treated as static (not traced).
        Default is ``()``.
    return_shape : bool, optional
        If True, the returned function produces a tuple ``(jaxpr, out_tree)``
        where ``out_tree`` is a PyTreeDef representing the structure of the
        output. This can be used to reconstruct PyTree objects from flat
        output lists using ``jax.tree_util.tree_unflatten``.
        Default is False.
    abstracted_axes : optional
        Specification for which axes to abstract over. Default is None.

    Returns
    -------
    Callable
        A function that, when called with example arguments, returns either:
        - A ClosedJaxpr representation of ``fun`` (if ``return_shape=False``)
        - A tuple ``(ClosedJaxpr, out_tree)`` (if ``return_shape=True``)

    Notes
    -----
    JAX's native ``make_jaxpr(return_shape=True)`` fails on custom abstract
    types because it tries to create ``ShapeDtypeStruct`` objects from the
    outputs, which requires ``shape`` and ``dtype`` attributes. This function
    avoids that by using ``jit(...).trace()`` to directly access the output
    tree structure.

    Examples
    --------
    >>> def f(x):
    ...     return {"a": x + 1, "b": x * 2}
    >>> jaxpr, out_tree = make_jaxpr_mod(f, return_shape=True)(1.0)
    >>> # out_tree can be used with tree_unflatten to reconstruct the dict
    """
    from jax._src.interpreters import partial_eval as pe
    from jax._src.util import split_list

    def jaxpr_creator(*args, **kwargs):
        if return_shape:
            # Use jit(...).trace() directly to get access to _out_tree
            # This avoids JAX's make_jaxpr return_shape logic which fails on
            # custom abstract types that don't have shape/dtype attributes.
            traced = jax.jit(
                fun, static_argnums=static_argnums, abstracted_axes=abstracted_axes
            ).trace(*args, **kwargs)

            # Extract the jaxpr, handling constants if needed
            # (same logic as JAX's make_jaxpr)
            if traced._num_consts:
                consts, _ = split_list(traced._args_flat, [traced._num_consts])
                jaxpr_ = pe.convert_invars_to_constvars(
                    traced.jaxpr.jaxpr, traced._num_consts
                )
                closed_jaxpr = ClosedJaxpr(jaxpr_, consts)
            else:
                closed_jaxpr = traced.jaxpr

            return closed_jaxpr, traced._out_tree
        else:
            return make_jaxpr(
                fun, static_argnums=static_argnums, abstracted_axes=abstracted_axes
            )(*args, **kwargs)

    return jaxpr_creator


def make_jaspr(fun, flatten_envs=True, return_shape=False, **jax_kwargs):
    """
    Creates a function that returns the Jaspr representation of a quantum function.

    This function is analogous to JAX's ``make_jaxpr``, but produces a Jaspr
    (a Jaxpr enhanced with quantum primitives) from a Qrisp quantum function.

    Parameters
    ----------
    fun : Callable
        The quantum function whose Jaspr is to be computed.
    flatten_envs : bool, optional
        If True (default), flatten quantum environments in the resulting Jaspr.
    return_shape : bool, optional
        If True, the returned function produces a tuple ``(jaspr, out_tree)``
        where ``out_tree`` is a PyTreeDef representing the structure of the
        output of ``fun``. This can be used to reconstruct PyTree objects
        from flat output lists using ``jax.tree_util.tree_unflatten``.
        Default is False.
    **jax_kwargs
        Additional keyword arguments passed to ``jax.make_jaxpr``, such as
        ``static_argnums``.

    Returns
    -------
    Callable
        A function that, when called with example arguments, returns either:
        - A Jaspr representation of ``fun`` (if ``return_shape=False``)
        - A tuple ``(Jaspr, out_tree)`` (if ``return_shape=True``) where ``out_tree`` is a PyTreeDef that can be used with ``tree_unflatten``

    Examples
    --------

    **Basic quantum circuit with measurement**

    Create a Jaspr for a simple Bell state circuit:

    ::

        from qrisp import QuantumVariable, h, cx, measure
        from qrisp.jasp import make_jaspr

        def simple_circuit():
            qv = QuantumVariable(2)
            h(qv[0])
            cx(qv[0], qv[1])
            return measure(qv)

        jaspr = make_jaspr(simple_circuit)()
        result = jaspr()  # Returns 0 or 3 with equal probability

    **Parameterized quantum circuit**

    Create a Jaspr with parameterized gates that can be executed with different parameters:

    ::

        from qrisp import QuantumVariable, h, p, measure
        from qrisp.jasp import make_jaspr

        def rotation_circuit(angle):
            qv = QuantumVariable(1)
            h(qv)
            p(angle, qv)
            return measure(qv)

        jaspr = make_jaspr(rotation_circuit)(0.5)
        result1 = jaspr(0.5)  # Execute with angle=0.5
        result2 = jaspr(1.0)  # Execute with angle=1.0

    **Using return_shape for PyTree reconstruction**

    Retrieve the output tree structure alongside the Jaspr for reconstructing
    complex return values:

    ::

        from qrisp import QuantumVariable, h, cx, x, measure
        from qrisp.jasp import make_jaspr
        from jax.tree_util import tree_unflatten, tree_flatten

        def multi_output_circuit():
            qa = QuantumVariable(2)
            qb = QuantumVariable(2)
            h(qa[0])
            cx(qa[0], qa[1])
            x(qb)
            return measure(qa), measure(qb)

        jaspr, out_tree = make_jaspr(multi_output_circuit, return_shape=True)()
        result_a, result_b = jaspr()

        # Use out_tree to reconstruct the output structure
        flat_results, _ = tree_flatten((result_a, result_b))
        reconstructed = tree_unflatten(out_tree, flat_results)

    """
    from qrisp import recursive_qv_search
    from qrisp.jasp import (
        AbstractQuantumState,
        TracingQuantumSession,
        check_for_tracing_mode,
    )

    # Handle static_argnums adjustment for the ammended function
    adjusted_jax_kwargs = dict(jax_kwargs)
    if "static_argnums" in adjusted_jax_kwargs:
        if isinstance(adjusted_jax_kwargs["static_argnums"], list):
            adjusted_jax_kwargs["static_argnums"] = list(
                adjusted_jax_kwargs["static_argnums"]
            )
            for i in range(len(adjusted_jax_kwargs["static_argnums"])):
                adjusted_jax_kwargs["static_argnums"][i] += 1
        else:
            adjusted_jax_kwargs["static_argnums"] += 1

    def jaspr_creator(*args, **kwargs):

        qs = TracingQuantumSession.get_instance()

        # Close any tracing quantum sessions that might have not been
        # properly closed due to whatever reason.
        if not check_for_tracing_mode():
            while qs.abs_qst is not None:
                qs.conclude_tracing()

        # This function will be traced by Jax.
        # Note that we add the abs_qst keyword as the tracing quantum circuit
        def ammended_function(*args, **kwargs):

            abs_qst = kwargs[10 * "~"]
            del kwargs[10 * "~"]

            qs.start_tracing(abs_qst)

            # If the signature contains QuantumVariables, these QuantumVariables went
            # through a flattening/unflattening procedure. The unflattening creates
            # a copy of the QuantumVariable object, which is however not yet registered in any
            # QuantumSession. We register these QuantumVariables in the current QuantumSession.
            arg_qvs = recursive_qv_search(args)
            for qv in arg_qvs:
                qs.register_qv(qv, None)

            try:
                res = fun(*args, **kwargs)
            except Exception as e:
                qs.conclude_tracing()
                raise e

            res_qc = qs.conclude_tracing()

            return res, res_qc

        ammended_kwargs = dict(kwargs)
        ammended_kwargs[10 * "~"] = AbstractQuantumState()

        # Use make_jaxpr_mod to trace the function
        # Pass return_shape through to get the output tree if requested
        static_argnums = adjusted_jax_kwargs.get("static_argnums", ())
        abstracted_axes = adjusted_jax_kwargs.get("abstracted_axes", None)

        result = make_jaxpr_mod(
            ammended_function,
            static_argnums=static_argnums,
            return_shape=return_shape,
            abstracted_axes=abstracted_axes,
        )(*args, **ammended_kwargs)

        if return_shape:
            closed_jaxpr, full_out_tree = result
            # full_out_tree is a PyTreeDef for (res, res_qc)
            # We use .children() to get the subtrees, and take the first one (res)
            user_out_tree = full_out_tree.children()[0]
        else:
            closed_jaxpr = result

        # Collect the environments
        # This means that the quantum environments no longer appear as
        # enter/exit primitives but as primitive that "call" a certain Jaspr.
        res = Jaspr.from_cache(collect_environments(closed_jaxpr))

        if flatten_envs:
            res = res.flatten_environments()

        if return_shape:
            return res, user_out_tree
        else:
            return res

    return jaspr_creator


def check_aval_equivalence(invars_1, invars_2):
    avals_1 = [invar.aval for invar in invars_1]
    avals_2 = [invar.aval for invar in invars_2]
    return all([type(avals_1[i]) == type(avals_2[i]) for i in range(len(avals_1))])


def remove_redundant_allocations(closed_jaxpr):
    """
    Optimizes the Jaspr by removing redundant qubit allocations.

    Strategy:
    1.  Map usages of all variables to identify how QubitArrays are consumed.
    2.  Identify `jasp.create_qubits` operations that are redundant. An allocation is redundant if:
        -   The resulting QubitArray is not returned by the function.
        -   The QubitArray is either unused (DropVar) or ONLY used by `free` primitives
            (like `get_size`, `delete_qubits`) that don't actually require the physical qubits
            if we know the allocation parameters.
    3.  Plan removals and replacements:
        -   Mark redundant `create_qubits` equations for removal.
        -   Mark dependent `get_size`, `reset`, `delete_qubits` equations for removal.
        -   Map the output circuit of removed equations to their input circuit (bypassing the operation).
        -   Map the output of `get_size` to the size input of the `create_qubits` equation.
    4.  Rewrite the Jaxpr:
        -   Filter out removed equations.
        -   Update variable references in remaining equations and output variables using the replacement map.
    """
    from jax.core import DropVar

    try:
        from jax.extend.core import Literal
    except ImportError:
        from jax.core import Literal

    jaxpr = closed_jaxpr.jaxpr
    eqns = jaxpr.eqns

    usages = {}

    # 1. Build usage map
    # We iterate over all equations and record which equation uses which input variable.
    for eqn in eqns:
        for var in eqn.invars:
            if isinstance(var, DropVar):
                continue
            if isinstance(var, Literal):
                continue
            if var not in usages:
                usages[var] = []
            usages[var].append(eqn)

    # Identify variables that are returned by the function, as these cannot be optimized away.
    returned_vars = set()
    for var in jaxpr.outvars:
        if not isinstance(var, DropVar) and not isinstance(var, Literal):
            returned_vars.add(var)

    replacements = {}
    eqns_to_remove = set()

    # Primitives that are "safe" to exist on a redundant qubit array.
    # These operations can be resolved without the actual qubits if the allocation is known.
    allowed_primitives = {"jasp.get_size", "jasp.delete_qubits"}

    # 2. Identify and plan removals
    for eqn in eqns:
        if eqn.primitive.name == "jasp.create_qubits":
            out_qa = eqn.outvars[0]
            out_qc = eqn.outvars[1]

            # If the QubitArray is part of the output, we must keep it.
            if out_qa in returned_vars:
                continue

            # Check if out_qa is only used by allowed primitives or not at all
            is_redundant = False

            if isinstance(out_qa, DropVar) or out_qa not in usages:
                is_redundant = True
            else:
                only_allowed_usages = True
                for u_eqn in usages[out_qa]:
                    if u_eqn.primitive.name not in allowed_primitives:
                        # Found a usage (e.g., gate) that requires real qubits.
                        only_allowed_usages = False
                        break
                if only_allowed_usages:
                    is_redundant = True

            if is_redundant:
                eqns_to_remove.add(id(eqn))

                # Rewire circuit: Bypass the allocation by mapping output circuit to input circuit.
                replacements[out_qc] = eqn.invars[1]

                # Handle dependent equations (safe primitives)
                if not isinstance(out_qa, DropVar) and out_qa in usages:
                    for u_eqn in usages[out_qa]:
                        eqns_to_remove.add(id(u_eqn))

                        if u_eqn.primitive.name == "jasp.get_size":
                            # Optimize get_size: Use the input size from create_qubits directly.
                            replacements[u_eqn.outvars[0]] = eqn.invars[0]
                        elif u_eqn.primitive.name == "jasp.delete_qubits":
                            # Bypass cleanup: Map the cleanup's output circuit to its input circuit.
                            replacements[u_eqn.outvars[0]] = u_eqn.invars[1]

    if not eqns_to_remove:
        return

    # 3. Apply replacements
    def resolve(var):
        # Literals are not hashable and cannot be in replacements
        if isinstance(var, Literal):
            return var
        try:
            # Resolving chains of replacements
            while var in replacements:
                var = replacements[var]
        except TypeError:
            # Handle other potentially unhashable types
            pass
        return var

    new_eqns = []
    for eqn in eqns:
        if id(eqn) in eqns_to_remove:
            continue

        # Update inputs of remaining equations using the replacement map
        eqn.invars[:] = [resolve(var) for var in eqn.invars]
        new_eqns.append(eqn)

    # Update the equations list in-place
    jaxpr.eqns[:] = new_eqns
    # Update the output variables of the jaxpr
    jaxpr.outvars[:] = [resolve(var) for var in jaxpr.outvars]



"""
********************************************************************************
* Copyright (c) 2026 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

# from __future__ import annotations
# from jax.typing import ArrayLike
from qrisp import (
    QuantumArray,
    QuantumVariable,
    QuantumBool,
    control,
    invert,
    rx,
    rz,
)
from qrisp.algorithms.gqsp.gqsp_angles import gqsp_angles
from qrisp.jasp import jrange
from typing import Any, Callable, Dict, Optional, Tuple, TYPE_CHECKING

if TYPE_CHECKING:
    from jax.typing import ArrayLike


# https://journals.aps.org/prxquantum/pdf/10.1103/PRXQuantum.5.020368
def GQSP(
    anc: QuantumBool,
    *qargs: QuantumVariable,
    unitary: Callable[..., None],
    p: Optional["ArrayLike"] = None,
    angles: Optional[Tuple["ArrayLike", "ArrayLike", "ArrayLike"]] = None,
    k: int = 0,
    kwargs: Dict[str, Any] = {},
) -> None:
    r"""
    Performs `Generalized Quantum Signal Processing <https://journals.aps.org/prxquantum/pdf/10.1103/PRXQuantum.5.020368>`_.

    Generalized Quantum Signal Processing was introduced by `Motlagh and Wiebe <https://journals.aps.org/prxquantum/pdf/10.1103/PRXQuantum.5.020368>`_. Its equivalence to
    the non-linear Fourier transform (NLFT) was subsequently established by `Laneve <https://arxiv.org/pdf/2503.03026>`_. 
    By adopting the conventions from Laneve, we treat Generalized Quantum Signal Processing (GQSP) as the overarching framework for single-qubit signal processing. 
    In this setting, previously known QSP variants—such as those constrained to $X$ or $Y$ rotations—emerge naturally as special cases where the underlying complex sequences of the NLFT are restricted to purely imaginary or real values, respectively.

    GQSP is described as follows:

    * Given a unitary $U$, 
    * and a complex degree $d$ polynomial $p(z)\in\mathbb C[z]$ such that $|p(e^{ix})|^2\leq 1$ for all $x\in\mathbb R$,

    this algorithm implements the unitary

    .. math::

        \begin{pmatrix} 
        p(U) & * \\ 
        * & *
        \end{pmatrix}
        &=R(\theta_0,\phi_0,\lambda)\left(\prod_{j=1}^{d}AR(\theta_j,\phi_j,0)\right)\\
        &=R(\theta_0,\phi_0,\lambda)AR(\theta_1,\phi_1,0)A\dotsc AR(\theta_d,\phi_d,0)

    where

    .. math::

        R(\theta,\phi,\lambda) = e^{i\lambda Z}e^{i\phi X}e^{i\theta Z} \in SU(2)

    and the angles $\theta,\phi\in\mathbb R^{d+1}$, $\lambda\in\mathbb R$ are calculated from the polynomial $p$, 
    $A=\begin{pmatrix}U & 0\\ 0 & I\end{pmatrix}$ is the signal operator.

    Parameters
    ----------
    anc : QuantumBool
        Auxiliary variable in state $\ket{0}$ for applying the GQSP protocol.
        Must be measured in state $\ket{0}$ for the GQSP protocol to be successful.
    *qargs : QuantumVariable
        QuantumVariables serving as operands for the unitary.
    unitary : Callable
        A function applying a unitary to the variables ``*qargs``.
        Typically, $U=e^{iH}$ for a Hermitian operator $H$ and GQSP applies a function of $H$.
    p : ArrayLike, optional
        1-D array containing the polynomial coefficients, ordered from lowest order term to highest.
        Either the polynomial ``p`` or ``angles`` must be specified.
    angles : tuple(ArrayLike, ArrayLike, ArrayLike), optional
        A tuple of angles $(\theta,\phi,\lambda)$ where $\theta,\phi\in\mathbb R^{d+1}$ are 1-D arrays
        and $\lambda\in\mathbb R$ is a scalar.
    k : int
        If specified, the Laurent polynomial $\tilde p(x)=x^{-k}p(x)$ is applied.
        The default is 0.
    kwargs : dict
        A dictionary of keyword arguments to pass to ``unitary``. The default is {}.

    Notes
    -----
    - The polynomial $p$ is rescaled automatically to satisfy $|p(e^{ix})|^2\leq 1$ for all $x\in\mathbb R$.
        
    Examples
    --------

    **Applying a transformation in Fourier basis**

    We apply the operator

    .. math::

        \cos(H) = \frac{e^{iH}+e^{-iH}}{2}

    for some :ref:`Hermitian operator <operators>` $H$ to the input state $\ket{\psi}=\ket{0}$.

    First, we define an operator $H$ and the unitary performing the Hamiltonian evolution $e^{iH}$.
    (In this case, Trotterization will perform Hamiltonian evolution exactly since the individual terms commute.)

    ::

        from qrisp import *
        from qrisp.gqsp import *
        from qrisp.operators import X,Y,Z
        import jax.numpy as jnp

        H = Z(0)*Z(1) + X(0)*X(1)

        def U(operand):
            H.trotterization(forward_evolution=False)(operand)


    Next, we define the ``operand_prep`` function that prepares a QuantumVariable is state $\ket{\psi}=\ket{0}$.

    ::

        def operand_prep():
            operand = QuantumVariable(2)
            return operand

    The transformation $\cos(H)$ is achieved by applying $\tilde p(x)=0.5x^{-1} + 0.5x^1$ to the unitary $e^{iH}$.
    This corresponds to the polynomial $p(x)=0.5+0.5x^2$ (i.e., ``p=[0.5,0,0.5]``) and ``k=1``. 

    Finally, we apply QSP within a :ref:`RUS` protocol.

    ::

        @RUS
        def inner():

            p = jnp.array([0.5,0,0.5])

            operand = operand_prep()
            anc = QuantumBool()
            GQSP(anc, operand, unitary=U, p=p, k=1)

            success_bool = measure(anc) == 0
            reset(anc)
            anc.delete()
            return success_bool, operand


        @terminal_sampling
        def main(): 

            qv = inner()
            return qv

    and simulate

    >>> main()
    {3: 0.85471756539818, 0: 0.14528243460182003}

    Let's compare to the classically calculated result:

    >>> A = H.to_array()
    >>> from scipy.linalg import cosm
    >>> print(cosm(A))
    [[ 0.29192658+0.j  0.        +0.j  0.        +0.j -0.70807342+0.j]
    [ 0.        +0.j  0.29192658+0.j  0.70807342+0.j  0.        +0.j]
    [ 0.        +0.j  0.70807342+0.j  0.29192658+0.j  0.        +0.j]
    [-0.70807342+0.j  0.        +0.j  0.        +0.j  0.29192658+0.j]]

    That is, starting in state $\ket{\psi}=\ket{0}=(1,0,0,0)$, we obtain

    >>> result = cosm(A)@(np.array([1,0,0,0]).transpose())
    >>> result = result/np.linalg.norm(result) # normalise
    >>> result = result**2 # compute measurement probabilities
    >>> print(result)
    [0.1452825+0.j 0.       +0.j 0.       +0.j 0.8547175-0.j]

    which are exactly the probabilities we observed in the quantum simulation.

    .. note::

        While GQSP allows you to apply arbitrary polynomials to operators, applying
        abitrary polynomials to :ref:`BlockEncodings <BlockEncoding>` requires an 
        additional step. This is because raising the operator

        .. math::

            U = \begin{pmatrix} \frac{A}{\alpha} & * \\ * & * \end{pmatrix}

        to a given power $k$ does not necessarily give you

        .. math::

            \tilde{U} = \begin{pmatrix} \left(\frac{A}{\alpha}\right)^k & * \\ * & * \end{pmatrix}

        In order to still apply polynomials also to them, we need to call the qubitization
        method and transform the polynomial into Chebychev basis. More to that in 
        the GQSP :ref:`tutorial`.

    """

    if angles is not None:
        theta, phi, lambda_ = angles
        d = len(theta) - 1
    elif p is not None:
        d = len(p) - 1
        (theta, phi, lambda_), _ = gqsp_angles(p)

    # Define R gate application function based on Theorem 9 in https://arxiv.org/abs/2503.03026
    def R(theta, phi, qubit):
        rz(-2 * theta, qubit)
        rx(-2 * phi, qubit)

    theta = theta[::-1]
    phi = phi[::-1]

    for i in jrange(d - k):
        R(theta[i], phi[i], anc)
        with control(anc, ctrl_state=0):
            unitary(*qargs, **kwargs)

    for i in jrange(k):
        R(theta[d - k + i], phi[d - k + i], anc)
        with control(anc, ctrl_state=1):
            with invert():
                unitary(*qargs, **kwargs)

    R(theta[d], phi[d], anc)
    rz(-2 * lambda_, anc)



"""
********************************************************************************
* Copyright (c) 2026 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import stim
from qrisp.circuit import Operation, QuantumCircuit


class StimNoiseGate(Operation):
    """
    Class for representing Stim errors in Qrisp circuits.

    This class is used to wrap Stim instructions into Qrisp Operations. These operations are
    effectively identity gates (they have an empty definition) but carry the information about the Stim
    noise channel. When converted to Stim circuits, these operations are replaced by the corresponding
    Stim instruction.

    .. list-table::
       :widths: 25 75
       :header-rows: 1

       * - Name
         - Description
       * - ``DEPOLARIZE1``
         - Single qubit depolarizing noise. This channel applies one of the Pauli errors X, Y, Z with probability $p/3$.
       * - ``DEPOLARIZE2``
         - Two qubit depolarizing noise. This channel applies one of the 15 non-identity two-qubit Pauli errors (IX, IY, ..., ZZ) with probability $p/15$.
       * - ``X_ERROR``
         - Single qubit Pauli-X error (Bit flip). Applies X with probability $p$.
       * - ``Y_ERROR``
         - Single qubit Pauli-Y error. Applies Y with probability $p$.
       * - ``Z_ERROR``
         - Single qubit Pauli-Z error (Phase flip). Applies Z with probability $p$.
       * - ``PAULI_CHANNEL_1``
         - Custom single qubit Pauli channel. Takes 3 arguments (px, py, pz) specifying the probabilities of applying X, Y, and Z errors respectively.
       * - ``PAULI_CHANNEL_2``
         - Custom two qubit Pauli channel. Takes 15 arguments specifying the probabilities of applying each of the 15 non-identity two-qubit Pauli errors.
       * - ``E`` (or ``CORRELATED_ERROR``)
         - Correlated Pauli error on multiple qubits (requires ``pauli_string`` argument). Applies the specified Pauli string with probability $p$.
       * - ``ELSE_CORRELATED_ERROR``
         - Similar to ``CORRELATED_ERROR`` but only applies if the *previous* error instruction did NOT apply an error. This allows constructing more complex conditional error models.


    Parameters
    ----------
    stim_name : str
        The name of the Stim error gate (e.g. ``DEPOLARIZE1``).
    *params : float
        The parameters of the error channel (e.g. error probability). Further
        details about the semantics of the parameters can be found in the
        `Stims gate reference <https://github.com/quantumlib/Stim/blob/main/doc/gates.md#noise-channels>`_
    pauli_string : str, optional
        A string of Pauli operators (e.g. ``XX``) for correlated errors.

    Examples
    --------

    We construct a simple circuit that contains both: quantum gates and error instructions.

    ::

        from qrisp import QuantumCircuit
        from qrisp.misc.stim_tools import StimNoiseGate
        qc = QuantumCircuit(1)
        qc.x(0)
        # Apply a depolarization error with probability 0.1
        qc.append(StimNoiseGate("DEPOLARIZE1", 0.1), qc.qubits)
        print(qc)
        # Yields:
        #       ┌───┐┌──────────────────┐
        # qb_0: ┤ X ├┤ stim.DEPOLARIZE1 ├
        #       └───┘└──────────────────┘
        print(qc.to_stim())
        # Yields:
        # X 0
        # DEPOLARIZE1(0.1) 0

    """

    def __init__(self, stim_name, *params, pauli_string=None):

        self.stim_name = stim_name
        self.pauli_string = pauli_string

        error_data = stim.gate_data(stim_name)

        parameter_amounts = list(error_data.num_parens_arguments_range)

        if not error_data.is_noisy_gate:
            raise Exception(
                "Non-noisy stim gates are not supported via the error interface. Please use the default Qrisp alternatives."
            )

        if len(params) not in parameter_amounts:
            raise Exception(
                f"Stim error of type {stim_name} can take parameter amounts {parameter_amounts} but received {len(params)} parameters instead"
            )

        if self.pauli_string is not None:
            # Check for compatibility
            if not (stim_name in ["E", "CORRELATED_ERROR", "ELSE_CORRELATED_ERROR"]):
                raise Exception(
                    f"Stim error {stim_name} does not support Pauli strings. Supported gates are E, CORRELATED_ERROR, ELSE_CORRELATED_ERROR"
                )

            num_qubits = len(self.pauli_string)

        elif error_data.is_single_qubit_gate:
            num_qubits = 1
        elif error_data.is_two_qubit_gate:
            num_qubits = 2
        else:
            raise Exception(
                f"Could not determine qubit amount for Stim error {stim_name}. Please check if the error is supported."
            )

        name = "stim." + stim_name

        Operation.__init__(
            self,
            name=name,
            num_qubits=num_qubits,
            params=params,
            definition=QuantumCircuit(num_qubits),
        )

        for k in self.permeability.keys():
            self.permeability[k] = False

    def inverse(self):
        raise Exception(f"StimNoiseGate {self.name} is not invertible.")

    def control(self, num_ctrl_qubits=1, ctrl_state=-1, method=None):
        raise Exception(f"StimNoiseGate {self.name} can not be controlled.")

    def c_if(self, num_control=1, ctrl_state=-1):
        raise Exception(f"StimNoiseGate {self.name} can not be classically controlled.")



"""
********************************************************************************
* Copyright (c) 2026 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

from typing import Callable

import jax.numpy as jnp
from jax import make_jaxpr
from jax.extend.core import ClosedJaxpr, Literal

from qrisp.jasp import check_for_tracing_mode


class ContextDict(dict):
    """
    Execution environment for the Jaxpr abstract interpreter.

    ``ContextDict`` is the central data structure used while evaluating a Jaxpr
    in the profiling/metric interpreters. It acts as a mutable environment that
    maps SSA-style Jaxpr variables (``eqn.invars`` / ``eqn.outvars``) to the
    concrete or traced values chosen by a metric to represent them.

    In a normal JAX execution, primitives are evaluated by the XLA runtime.
    In the profiler, however, we *replay* the Jaxpr equation-by-equation in
    pure Python using a custom ``eqn_evaluator``. During this replay, every
    Jaxpr variable must be associated with a value. ``ContextDict`` provides
    this association.

    Conceptually, this is equivalent to the “environment” of an interpreter:

        Jaxpr variable  --->  metric-specific runtime value

    The values stored in the dictionary are **not** the original program values.
    Instead, they are arbitrary Python/JAX objects chosen by the active metric
    to represent the meaning of that variable for the purpose of resource
    analysis. For example:

    - In ``count_ops``:
        * ``QuantumCircuit``  ->  (counting_array, incrementation_constants)
        * ``QubitArray``      ->  integer size
        * ``Qubit``           ->  None
        ...

    - In ``depth``:
        * ``QuantumCircuit``  ->  (depth_array, current_depth, invalid_flag)
        * ``QubitArray``      ->  (lookup_table, logical_size)
        * ``Qubit``           ->  integer qubit id
        ...

    and so on for other metrics.

    The Jaxpr structure itself is **never modified**. Only the interpretation
    of its variables changes through what is stored in this dictionary.

    Special behavior
    ----------------
    ``ContextDict`` overrides ``__getitem__`` for two main reasons:

    1. **Literal handling**

       Jaxpr literals (``jax.core.Literal``) do not appear as normal variables.
       When accessed, they are transparently converted to their Python value.

    2. **Automatic JAX scalar conversion**

       If a stored value is a Python ``int`` or ``float``, it is automatically
       converted to a JAX scalar array. This guarantees that subsequent JAX
       operations inside the metric evaluator can treat these values as valid
       JAX types, avoiding abstractification errors.

    Because metrics are free to choose what values to associate with variables,
    this mechanism allows multiple resource metrics (gate counts, depth, future
    metrics) to reuse the same Jaxpr interpreter without changing its logic —
    only the contents of this dictionary differ.

    In summary, ``ContextDict`` is the bridge between:

        static Jaxpr structure  <-->  dynamic metric-specific semantics
    """

    def __getitem__(self, key):
        """Override to handle Jaxpr literals and automatic JAX scalar conversion."""

        if isinstance(key, Literal):
            res = key.val
        else:
            res = dict.__getitem__(self, key)

        if type(res) == int:
            return jnp.asarray(res, dtype=jnp.dtype("int64"))
        if type(res) == float:
            return jnp.asarray(res, dtype=jnp.dtype("float64"))
        return res


def exec_eqn(eqn, context_dic):
    """Evaluate a single equation within the given context dictionary."""

    invalues = extract_invalues(eqn, context_dic)
    res = eqn.primitive.bind(*invalues, **eqn.params)
    insert_outvalues(eqn, context_dic, res)


def eval_jaxpr(jaxpr, return_context_dic=False, eqn_evaluator=exec_eqn) -> Callable:
    """
    Evaluates a Jaxpr using the provided equation evaluator.

    Parameters
    ----------
    jaxpr : jax.core.Jaxpr | ClosedJaxpr | Jaspr
        The JAX expression to evaluate.

    return_context_dic : bool, optional
        Whether to return the context dictionary along with the output values,
        by default False.

    eqn_evaluator : Callable, optional
        The function to evaluate each equation in the JAX expression,
        by default exec_eqn.

    Returns
    -------
    Callable
        A function that evaluates the jaxpr.

    """

    # Import here to avoid circular imports
    from qrisp.jasp.jasp_expression import Jaspr

    consts = []
    if isinstance(jaxpr, ClosedJaxpr):
        consts = list(jaxpr.consts)
        jaxpr_core = jaxpr.jaxpr
    elif isinstance(jaxpr, Jaspr):
        consts = list(jaxpr.consts)
        jaxpr_core = jaxpr
    else:
        jaxpr_core = jaxpr

    invars = list(jaxpr_core.constvars) + list(jaxpr_core.invars)
    outvars = list(jaxpr_core.outvars)

    def jaxpr_evaluator(*args):

        args_and_consts = consts + list(args)

        if len(args_and_consts) != len(invars):
            raise ValueError(
                "Tried to evaluate jaxpr with insufficient arguments: "
                f"expected {len(invars)} (including {len(consts)} consts), got {len(args_and_consts)}."
            )

        context_dic = ContextDict(dict(zip(invars, args_and_consts)))

        eval_jaxpr_with_context_dic(jaxpr_core, context_dic, eqn_evaluator)

        if return_context_dic:
            outputs = [context_dic] + [context_dic[v] for v in outvars]
        else:
            outputs = [context_dic[v] for v in outvars]

        return outputs[0] if len(outputs) == 1 else tuple(outputs)

    return jaxpr_evaluator


def reinterpret(jaxpr, eqn_evaluator=exec_eqn):

    if isinstance(jaxpr, ClosedJaxpr):
        inter_jaxpr = jaxpr.jaxpr
    else:
        inter_jaxpr = jaxpr

    res = make_jaxpr(eval_jaxpr(inter_jaxpr, eqn_evaluator=eqn_evaluator))(
        *[var.aval for var in inter_jaxpr.constvars + inter_jaxpr.invars]
    ).jaxpr

    res.constvars.extend(res.invars[: len(inter_jaxpr.constvars)])
    temp = list(res.invars[len(inter_jaxpr.constvars) :])
    res.invars.clear()
    res.invars.extend(temp)

    if isinstance(jaxpr, ClosedJaxpr):
        res = ClosedJaxpr(res, jaxpr.consts)

    return res


def eval_jaxpr_with_context_dic(jaxpr, context_dic, eqn_evaluator=exec_eqn):

    for eqn in jaxpr.eqns:

        default_eval = eqn_evaluator(eqn, context_dic)

        if default_eval:
            if (
                eqn.primitive.name in ["while", "cond", "scan"]
                and not check_for_tracing_mode()
            ):

                from qrisp.jasp import (
                    evaluate_cond_eqn,
                    evaluate_scan,
                    evaluate_while_loop,
                )

                if eqn.primitive.name == "while":
                    evaluate_while_loop(eqn, context_dic, eqn_evaluator)
                elif eqn.primitive.name == "cond":
                    evaluate_cond_eqn(eqn, context_dic, eqn_evaluator)
                else:
                    evaluate_scan(eqn, context_dic, eqn_evaluator)

                continue

            exec_eqn(eqn, context_dic)


def extract_invalues(eqn, context_dic):
    invalues = []
    for i in range(len(eqn.invars)):
        invar = eqn.invars[i]
        invalues.append(context_dic[invar])
    return invalues


def extract_constvalues(eqn, context_dic):
    constvalues = []
    for i in range(len(eqn.constvars)):
        constvar = eqn.constvars[i]
        constvalues.append(context_dic[constvar])

    return constvalues


def insert_outvalues(eqn, context_dic, outvalues):

    if eqn.primitive.multiple_results:
        if len(outvalues) != len(eqn.outvars):
            raise Exception(
                "Tried to insert invalid amount of values into the Context Dictionary"
            )

        for i in range(len(eqn.outvars)):
            context_dic[eqn.outvars[i]] = outvalues[i]
    else:
        context_dic[eqn.outvars[0]] = outvalues



"""
********************************************************************************
* Copyright (c) 2026 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import numpy as np
import jax.numpy as jnp
from qrisp import QuantumBool
from qrisp.core.gate_application_functions import rx
from qrisp.environments import conjugate, control
from qrisp.alg_primitives.reflection import reflection
from qrisp.algorithms.gqsp.gqsp_angles import gqsp_angles
from qrisp.algorithms.gqsp.helper_functions import poly2cheb, _rescale_poly
from qrisp.block_encodings import BlockEncoding
from qrisp.jasp import jrange
from qrisp.operators import QubitOperator, FermionicOperator
from typing import Literal, TYPE_CHECKING

if TYPE_CHECKING:
    from jax.typing import ArrayLike


def QET(
    H: BlockEncoding | FermionicOperator | QubitOperator,
    p: "ArrayLike",
    kind: Literal["Polynomial", "Chebyshev"] = "Polynomial",
    rescale: bool = True,
) -> BlockEncoding:
    r"""
    Returns a BlockEncoding representing a polynomial transformation of the operator via `Quantum Eigenvalue Transform <https://arxiv.org/pdf/2312.00723>`_.

    For a block-encoded **Hermitian** operator $H$ and a **real, fixed parity** polynomial $p(x)$, this method returns
    a BlockEncoding of the operator $p(H)$.

    The Quantum Eigenvalue Transform is described as follows:

    * Given a Hermitian operator $H=\sum_i\lambda_i\ket{\lambda_i}\bra{\lambda_i}$ where $\lambda_i\in\mathbb R$ are the eigenvalues for the eigenstates $\ket{\lambda_i}$,
    * A quantum state $\ket{\psi}=\sum_i\alpha_i\ket{\lambda_i}$ where $\alpha_i\in\mathbb C$ are the amplitudes for the eigenstates $\ket{\lambda_i}$,
    * A (complex) polynomial $p(z)$,

    this transformation prepares a state proportional to

    .. math::

        p(H)\ket{\psi}=\sum_i p(\lambda_i)\ket{\lambda_i}\bra{\lambda_i}\sum_j\alpha_j\ket{\lambda_j}=\sum_i p(\lambda_i)\alpha_i\ket{\lambda_i}

    Parameters
    ----------
    H : BlockEncoding | FermionicOperator | QubitOperator
        The Hermitian operator to be transformed.
    p : ArrayLike
        1-D array containing the polynomial coefficients, ordered from lowest order term to highest.
    kind : {"Polynomial", "Chebyshev"}
        The basis in which the coefficients are defined.

        - ``"Polynomial"``: $p(x) = \sum c_i x^i$

        - ``"Chebyshev"``: $p(x) = \sum c_i T_i(x)$, where $T_i$ are Chebyshev polynomials of the first kind.

        Default is ``"Polynomial"``.
    rescale : bool
        If True (default), the method returns a block-encoding of $p(H)$.
        If False, the method returns a block-encoding of $p(H/\alpha)$ where $\alpha$ is the normalization factor for the block-encoding of the operator $H$.

    Returns
    -------
    BlockEncoding
        A new BlockEncoding instance representing the transformed operator $p(H)$.

    Notes
    -----
    - Improved efficiency compared to GQET for a real, fixed parity polynomial $p(x)$.

    Examples
    --------

    Define a Hermitian matrix $A$ and a vector $\vec{b}$.

    ::

        import numpy as np

        A = np.array([[0.73255474, 0.14516978, -0.14510851, -0.0391581],
                    [0.14516978, 0.68701415, -0.04929867, -0.00999921],
                    [-0.14510851, -0.04929867, 0.76587818, -0.03420339],
                    [-0.0391581, -0.00999921, -0.03420339, 0.58862043]])

        b = np.array([0, 1, 1, 1])

    Generate a BlockEncoding of $A$ and use QET to obtain a BlockEncoding of $p(H)$
    for a real polynomial of even parity.

    ::

        from qrisp import *
        from qrisp.block_encodings import BlockEncoding
        from qrisp.gqsp import QET

        BA = BlockEncoding.from_array(A)

        BA_poly = QET(BA, np.array([1.,0.,1.]))

        # Prepares operand variable in state |b>
        def prep_b():
            operand = QuantumVariable(2)
            prepare(operand, b)
            return operand

        @terminal_sampling
        def main():
            operand = BA_poly.apply_rus(prep_b)()
            return operand

        res_dict = main()
        amps = np.sqrt([res_dict.get(i, 0) for i in range(len(b))])

    Finally, compare the quantum simulation result with the classical solution:

    ::

        c = (np.eye(4) + A @ A) @ b
        c = c / np.linalg.norm(c)

        print("QUANTUM SIMULATION\n", amps, "\nCLASSICAL SOLUTION\n", c)
        #QUANTUM SIMULATION
        # [0.02405799 0.57657687 0.61493257 0.53743675]
        #CLASSICAL SOLUTION
        # [-0.024058    0.57657692  0.61493253  0.53743673]

    """

    # is_real = jnp.allclose(p.imag, 0, atol=1e-6)
    # is_even = jnp.allclose(p[1::2], 0, atol=1e-6)
    is_odd = jnp.allclose(p[0::2], 0, atol=1e-6)

    ALLOWED_KINDS = {"Polynomial", "Chebyshev"}
    if kind not in ALLOWED_KINDS:
        raise ValueError(
            f"Invalid kind specified: '{kind}'. "
            f"Allowed kinds are: {', '.join(ALLOWED_KINDS)}"
        )

    if isinstance(H, (QubitOperator, FermionicOperator)):
        H = BlockEncoding.from_operator(H)

    # Rescaling of the polynomial to account for scaling factor alpha of block-encoding
    if rescale:
        p = _rescale_poly(H.alpha, p, kind=kind)
    if kind == "Polynomial":
        p = poly2cheb(p)

    m = len(H._anc_templates)
    d = len(p)
    # Angles theta and lambda vanish for real polynomials https://arxiv.org/abs/2503.03026.
    # Implementation based on conjecture: phi has fixed parity iff p has fixed parity.
    # Combine two consecutive walk operators: (R U) (R U) = (R U R U ) = (R U_dg R U) = T_2
    # This is qubitization step even if the block encoding unitary is not Hemitian.
    # See https://math.berkeley.edu/~linlin/qasc/qasc_notes.pdf (page 104).
    # If the parity is odd, there is a single (R U) at the end which is not followed by a rotation.
    # Since the QET is only successful if all ancillas are |0>, there is no need to control-(R U)
    # and the refelction acts as identity.

    angles, alpha = gqsp_angles(p)
    phi = angles[1][::-1]

    def T2(*args):
        with conjugate(H.unitary)(*args):
            reflection(args[:m])
        reflection(args[:m])

    def new_unitary(*args):

        rx(-2 * phi[0], args[0])
        for i in jrange(1, (d + 1) // 2):
            with control(args[0], ctrl_state=0):
                T2(*args[1:])
            rx(-2 * phi[2 * i], args[0])

        with control(is_odd):
            H.unitary(*args[1:])

    new_anc_templates = [QuantumBool().template()] + H._anc_templates
    return BlockEncoding(
        alpha, new_anc_templates, new_unitary, num_ops=H.num_ops, is_hermitian=False
    )



"""
********************************************************************************
* Copyright (c) 2026 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import numpy as np
from scipy.special import comb
import jax.numpy as jnp
from qrisp import (
    QuantumVariable,
    QuantumFloat,
    x,
    z,
    ry,
    control,
    conjugate,
    invert,
)
from qrisp.alg_primitives.reflection import reflection
from qrisp.block_encodings import BlockEncoding
from qrisp.jasp import jrange
import numpy.typing as npt
from typing import Tuple, TYPE_CHECKING

if TYPE_CHECKING:
    from jax.typing import ArrayLike


def cks_params(eps: float, kappa: float, max_beta: int = None) -> Tuple[int, int]:
    """
    Computes the complexity parameter :math:`\\beta` and the truncation order :math:`j_0` for the
    truncated Chebyshev approximation of :math:`1/x`, as described in the `Childs–Kothari–Somma paper <https://arxiv.org/abs/1511.02306>`_.

    Given the condition number :math:`\\kappa` and the target precision
    :math:`\\epsilon`, the parameters are computed as:

    .. math::

        \\beta = \kappa^2 \log\!\left(\\frac{\kappa}{\epsilon}\\right),
        \quad
        j_0 = \sqrt{\\beta \log\!\left(\\frac{4\\beta}{\epsilon}\\right)}.

    If ``max_beta`` is provided, :math:`\\beta` is capped to
    :math:`\\min(\\beta, \\beta_{\\max})`. The returned values are cast to integers
    via the floor operation.

    Parameters
    ----------
    eps : float
        Target precision :math:`\epsilon`.
    kappa : float
        An upper bound for the condition number :math:`\\kappa` of :math:`A`. This value defines the "gap"
        around zero where the function :math:`1/x` is not approximated.
    max_beta : float, optional
        Optional upper bound on the complexity parameter :math:`\\beta`.

    Returns
    -------
    j0 : int
        Truncation order of the Chebyshev expansion
        :math:`j_0 = \\lfloor\sqrt{\\beta \log(4\\beta/\epsilon)}\\rfloor`.
    beta : int
        Complexity parameter :math:`\\beta = \\lfloor\kappa^2 \log(\kappa/\epsilon)\\rfloor`.
    """

    if max_beta == None:
        beta = kappa**2 * np.log(kappa / eps)
    else:
        beta = min(kappa**2 * np.log(kappa / eps), max_beta)
    j0 = np.sqrt(beta * np.log(4 * beta / eps))
    return int(j0), int(beta)


def cks_coeffs(j0: int, b: int) -> npt.NDArray[float]:
    """
    Computes the positive coefficients :math:`\\alpha_i` for the truncated
    Chebyshev expansion of :math:`1/x` up to order :math:`2j_0+1`.

    The approximation is expressed as a linear combination
    of odd Chebyshev polynomials truncated at index :math:`j_0` (Lemma 14):

    .. math::

        g(x) = 4 \\sum_{j=0}^{j_0} (-1)^j
        \\left[ \\sum_{i=j+1}^{b} \\frac{\\binom{2b}{b+i}}{2^{2b}} \\right]
        T_{2j+1}(x)

    The Linear Combination of Unitaries (LCU) lemma requires strictly
    positive coefficients :math:`\\alpha_i > 0`, their absolute values are
    used. The alternating factor :math:`(-1)^j` is later implemented as a set of
    Z-gates within the CKS circuit.

    Parameters
    ----------
    j0 : int
        Truncation order of the Chebyshev expansion,
        :math:`j_0 = \\lfloor\sqrt{\\beta \log(4\\beta/\epsilon)}\\rfloor`.
    b : float
        Complexity parameter :math:`\\beta = \\lfloor\kappa^2 \log(\kappa/\epsilon)\\rfloor`.

    Returns
    -------
    coeffs : ndarray
        1-D array of positive Chebyshev coefficients :math:`{\\alpha_{2j+1}}`,
        corresponding to the odd degree Chebyshev polynomials of the first kind :math:`T_1, T_3, \\dotsc, T_{2j_0+1}`.

    """
    coeffs = []
    for j in range(j0 + 1):
        sum_i = 0
        for i in range(j + 1, b + 1):
            sum_i += comb(2 * b, b + i)

        coeff = 4 * (2 ** (-2 * b)) * sum_i
        coeffs.append(coeff)
    return np.array(coeffs)


def _unary_angles(coeffs: "ArrayLike") -> "ArrayLike":
    """
    Computes rotation angles :math:`\\phi_i` to prepare the unary state :math:`\ket{\\text{unary}}`,
    corresponding to the square-root-amplitude encoding of the Chebyshev coefficients.

    The prepared unary state is

    .. math::

        \ket{\\text{unary}} \propto
        \\sqrt{\\alpha_1}\ket{100\\dots00} + \\sqrt{\\alpha_3}\ket{110\\dots00} +
        \\cdots + \\sqrt{\\alpha_{2j_0+1}}\ket{111\\dots11}

    Parameters
    ----------
    coeffs : ArrayLike
        1-D array of positive Chebyshev coefficients :math:`\\alpha_i`.

    Returns
    -------
    ArrayLike
        1-D array of rotation angles :math:`\\phi_i` for unary state preparation.
    """

    alpha = jnp.sqrt(
        coeffs
    )  # coeffs need not to be normalized since unary angles only depend on their ratio
    phi = jnp.zeros(len(alpha) - 1)
    phi = phi.at[-1].set(
        jnp.arctan(alpha[-1] / alpha[-2])
    )  # Last angle is determined by ratio of final two amplitudes
    for i in range(len(phi) - 2, -1, -1):
        phi = phi.at[i].set(
            jnp.arctan(alpha[i + 1] / alpha[i] / jnp.cos(phi[i + 1]))
        )  # Recursively compute previous rotation angles
    return 2 * phi  # Compensate for factor 1/2 in ry


def unary_prep(case: QuantumVariable, coeffs: "ArrayLike") -> None:
    """
    Prepares the unary-encoded state :math:`\ket{\\text{unary}}` of Chebyshev coefficients.

    When applied to a variable in state :math:`\ket{0}`, the resulting superposition is

    .. math::

        \ket{\\text{unary}} \propto
        \\sqrt{\\alpha_1}\ket{100\\dots00} + \\sqrt{\\alpha_3}\ket{110\\dots00} +
        \\cdots + \\sqrt{\\alpha_{2j_0+1}}\ket{111\\dots11}

    Parameters
    ----------
    case : QuantumVariable
        Variable with :math:`j_0` qubits on which the unary state preparation will be performed.
        If the variable is in state :math:`\ket{0}`, the state :math:`\ket{\\text{unary}}` is prepared.
    coeffs : ArrayLike
        1-D array of :math:`j_0` Chebyshev coefficients :math:`\\alpha_1,\\alpha_3,\dotsc,\\alpha_{2j_0+1}`.

    """
    phi = _unary_angles(coeffs)

    x(case[0])
    ry(phi[0], case[1])
    for i in jrange(1, case.size - 1):
        with control(case[i]):
            ry(phi[i], case[i + 1])


def CKS(
    A: BlockEncoding, eps: float, kappa: float, max_beta: float = None
) -> BlockEncoding:
    """
    Performs the `Childs–Kothari–Somma (CKS) quantum algorithm <https://arxiv.org/abs/1511.02306>`_ to solve the Quantum Linear System Problem (QLSP)
    :math:`A \\vec{x} = \\vec{b}`, using the Chebyshev approximation of :math:`1/x`.
    When applied to a state $\ket{b}$, the algorithm prepares a state :math:`\ket{\\tilde{x}} \propto A^{-1} \ket{b}`
    within target precision :math:`\epsilon` of the ideal solution :math:`\ket{x}`.

    For a block-encoded **Hermitian** matrix :math:`A`, this function returns a BlockEncoding of an
    operator :math:`\\tilde{A}^{-1}` such that :math:`\|\\tilde{A}^{-1} - A^{-1}\| \leq \epsilon`.
    The inversion is implemented using a polynomial approximation of :math:`1/x` over the domain :math:`D_{\\kappa} = [-1, -1/\\kappa] \\cup [1/\\kappa, 1]`.

    The asymptotic complexity is
    :math:`\\mathcal{O}\\!\\left(\\log(N)s \\kappa^2 \\text{polylog}\\!\\frac{s\\kappa}{\\epsilon}\\right)`, where :math:`N` is the matrix size, :math:`s` its sparsity, and
    :math:`\kappa` its condition number. This represents an exponentially
    better precision scaling compared to the HHL algorithm.

    This function integrates all core components of the CKS approach:
    Chebyshev polynomial approximation, Linear Combination of Unitaries (LCU),
    `qubitization with reflection operators <https://arxiv.org/abs/2208.00567>`_, and the :ref:`Repeat-Until-Success (RUS) <RUS>` protocol.
    The semantics of the approach can be illustrated with the following circuit schematics:

    .. image:: /_static/CKS_schematics.png
       :align: center

    Implementation overview:
      1. Compute the CKS parameters :math:`j_0` and :math:`\\beta` (:func:`cks_params`).
      2. Generate Chebyshev coefficients and the auxiliary unary state (:func:`cks_coeffs`, :func:`unary_prep`).
      3. Build the core LCU structure via qubitization operator.

    The goal of this algorithm is to apply the non-unitary operator :math:`A^{-1}` to the
    input state :math:`\ket{b}`. Following the Chebyshev approach introduced in the CKS paper, we express :math:`A^{-1}` as
    a linear combination of odd Chebyshev polynomials:

    .. math::

        A^{-1}\propto\sum_{j=0}^{j_0}\\alpha_{2j+1}T_{2j+1}(A),

    where :math:`T_k(A)` are Chebyshev polynomials of the first kind and :math:`\\alpha_{2j+1} > 0` are computed
    via :func:`cks_coeffs`. These operators can be efficiently implemented with qubitization, which relies on a unitary
    block encoding :math:`U` of the matrix :math:`A`, and a :ref:`reflection operator <reflection>` :math:`R`.

    If the block encoding unitary is $U$ is Hermitian (:math:`U^2=I`),
    the fundamental iteration step is defined as :math:`(RU)`, where :math:`R` reflects
    around the auxiliary block-encoding state :math:`\ket{G}`, prepared as the ``inner_case`` QuantumFloat.
    Repeated applications of these unitaries, :math:`(RU)^k`, yield a block encoding of the :math:`k`-th Chebyshev polynomial
    of the first kind :math:`T_k(A)`.

    We then construct a linear combination of these block-encoded polynomials using the LCU structure

    .. math::

       \\text{LCU}\ket{0}\ket{\psi}=\\text{PREP}^{\dagger}\cdot \\text{SEL}\cdot \\text{PREP}\ket{0}\ket{\psi}=\\tilde{A}\ket{0}\ket{\psi}.

    Here, the :math:`\\text{PREP}` operation prepares an auxiliary ``out_case`` Quantumfloat in the unary state :math:`\ket{\\text{unary}}`
    that encodes the square root of the Chebyshev coefficients :math:`\sqrt{\\alpha_j}`. The :math:`\\text{SEL}` operation selects and applies the
    appropriate Chebyshev polynomial operator :math:`T_k(A)`, implemented by :math:`(RU)^k`, controlled on ``out_case`` in the unary
    state :math:`\ket{\\text{unary}}`. Based on the Hamming-weight :math:`k` of :math:`\ket{\\text{unary}}`,
    the polynomial :math:`T_{2k-1}` is block encoded and applied to the circuit.

    To construct a linear combination of Chebyshev polynomials up to the :math:`2j_0+1`-th order, as in the original paper,
    our implementation requires :math:`j_0+1` qubits in the ``out_case`` state :math:`\ket{\\text{unary}}`.

    The Chebyshev coefficients alternate in sign :math:`(-1)^j\\alpha_j`.
    Since the LCU lemma requires :math:`\\alpha_j>0`, negative terms are accounted for
    by applying Z-gates on each index qubit in ``out_case``.

    Parameters
    ----------
    A : BlockEncoding
        The block-encoded Hermitian matrix to be inverted.
    eps : float
        Target precision :math:`\epsilon`.
    kappa : float
        An upper bound for the condition number :math:`\\kappa` of :math:`A`. This value defines the "gap"
        around zero where the function :math:`1/x` is not approximated.
    max_beta : float, optional
        Optional upper bound on the complexity parameter :math:`\\beta`.

    Returns
    -------
    BlockEncoding
        A new BlockEncoding instance representing an approximation of the inverse $A^{-1}$.

    Examples
    --------

    The following examples demonstrate how the CKS algorithm can be applied to solve the
    quantum linear systems problem (QLSP)
    :math:`A \\vec{x} = \\vec{b}`, using either a direct Hermitian matrix input or
    a preconstructed block-encoding representation.

    **Example 1: Solving a 4×4 Hermitian system**

    First, we define a small Hermitian matrix :math:`A` and a right-hand side vector :math:`\\vec{b}`:

    ::

        import numpy as np

        A = np.array([[0.73255474, 0.14516978, -0.14510851, -0.0391581],
                      [0.14516978, 0.68701415, -0.04929867, -0.00999921],
                      [-0.14510851, -0.04929867, 0.76587818, -0.03420339],
                      [-0.0391581, -0.00999921, -0.03420339, 0.58862043]])

        b = np.array([0, 1, 1, 1])

        kappa = np.linalg.cond(A)
        print("Condition number of A: ", kappa)
        # Condition number of A:  1.8448536035491883

    Next, we solve this linear system using the CKS quantum algorithm:

    ::

        from qrisp import prepare, QuantumFloat
        from qrisp.algorithms.cks import CKS
        from qrisp.block_encodings import BlockEncoding
        from qrisp.jasp import terminal_sampling

        def prep_b():
            operand = QuantumFloat(2)
            prepare(operand, b)
            return operand

        @terminal_sampling
        def main():

            BA = BlockEncoding.from_array(A)
            x = CKS(BA, 0.01, kappa).apply_rus(prep_b)()
            return x

        res_dict = main()

        # The resulting dictionary contains the measurement probabilities
        # for each computational basis state.
        # To extract the corresponding quantum amplitudes (up to sign):
        amps = np.sqrt([res_dict.get(i, 0) for i in range(len(b))])
        print("QUANTUM SIMULATION\\n", amps)
        # QUANTUM SIMULATION
        # [0.02714082 0.55709921 0.53035395 0.63845794]

    We compare this to the classical normalized solution:

    ::

        c = (np.linalg.inv(A) @ b) / np.linalg.norm(np.linalg.inv(A) @ b)

        print("CLASSICAL SOLUTION\\n", c)
        # CLASSICAL SOLUTION
        # [0.02944539 0.55423278 0.53013239 0.64102936]

    We see that we obtained the same result in our quantum simulation up to precision :math:`\epsilon`!

    To perform quantum resource estimation, replace the ``@terminal_sampling``
    decorator with ``@count_ops(meas_behavior="0")``:

    ::

        from qrisp.jasp import count_ops

        @count_ops(meas_behavior="0")
        def main():

            BA = BlockEncoding.from_array(A)
            x = CKS(BA, 0.01, kappa).apply_rus(prep_b)()
            return x

        res_dict = main()
        print(res_dict)
        # {'gphase': 51, 't_dg': 1548, 'cz': 150, 'ry': 2, 'z': 12, 'h': 890, 'cy': 50, 'cx': 3516, 'x': 377, 's_dg': 70, 'p': 193, 's': 70, 't': 1173, 'u3': 753, 'measure': 18}

    The printed dictionary lists the estimated quantum gate counts required for the CKS algorithm. Since the simulation itself is not executed,
    this approach enables scalable gate count estimation for linear systems of arbitrary size.

    **Example 2: Using a custom block encoding**

    The previous example displays how to solve the linear system for any Hermitian matrix :math:`A`,
    where the block-encoding is constructed from the Pauli decomposition of :math:`A`. While this approach is efficient when :math:`A`
    corresponds to, e.g., an Ising or a Heisenberg Hamiltonian, the number of Pauli terms may not scale favorably in general.
    For certain sparse matrices, their structure can be exploited to construct more efficient block-encodings.

    In this example, we construct a block-encoding representation for the following following tridiagonal sparse matrix:

    ::

        import numpy as np

        def tridiagonal_shifted(n, mu=1.0, dtype=float):
            I = np.eye(n, dtype=dtype)
            return (2 + mu) * I - 2*np.eye(n, k=n//2, dtype=dtype) - 2*np.eye(n, k=-n//2, dtype=dtype)

        N = 8
        A = tridiagonal_shifted(N, mu=3)
        b = np.random.randint(0, 2, size=N)

        print(A)
        # [[ 5.  0.  0.  0. -2.  0.  0.  0.]
        #  [ 0.  5.  0.  0.  0. -2.  0.  0.]
        #  [ 0.  0.  5.  0.  0.  0. -2.  0.]
        #  [ 0.  0.  0.  5.  0.  0.  0. -2.]
        #  [-2.  0.  0.  0.  5.  0.  0.  0.]
        #  [ 0. -2.  0.  0.  0.  5.  0.  0.]
        #  [ 0.  0. -2.  0.  0.  0.  5.  0.]
        #  [ 0.  0.  0. -2.  0.  0.  0.  5.]]

    This matrix can be decomposed using three unitaries: the identity $I$, and two shift operators $V\colon\ket{k}\\rightarrow-\ket{k+N/2 \mod N}$ and $V^{\dagger}\colon\ket{k}\\rightarrow-\ket{k-N/2 \mod N}$.
    We define their corresponding functions and the block-encoding using Linear Combination of Unitaries (LCU):

    ::

        from qrisp import gphase

        def I(qv):
            pass

        def V(qv):
            qv += N//2
            gphase(np.pi, qv[0])

        def V_dg(qv):
            qv -= N//2
            gphase(np.pi, qv[0])

        unitaries = [I, V, V_dg]
        coeffs = np.array([5,1,1,0])
        BA = BlockEncoding.from_lcu(coeffs, unitaries)

    Next, we solve this linear system using the CKS quantum algorithm:

    ::

        def prep_b():
            operand = QuantumFloat(3)
            prepare(operand, b)
            return operand

        @terminal_sampling
        def main():

            x = CKS(BA, 0.01, kappa).apply_rus(prep_b)()
            return x

        res_dict = main()
        amps = np.sqrt([res_dict.get(i, 0) for i in range(len(b))])

    We, again, compare the solution obtained by quantum simulation with the classical solution

    ::

        c = (np.linalg.inv(A) @ b) / np.linalg.norm(np.linalg.inv(A) @ b)

        print("QUANTUM SIMULATION\\n", amps, "\\nCLASSICAL SOLUTION\\n", c)
        # QUANTUM SIMULATION
        # [0.43917486 0.31395935 0.12522139 0.43917505 0.43917459 0.12522104 0.31395943 0.43917502]
        # CLASSICAL SOLUTION
        # [0.43921906 0.3137279  0.12549116 0.43921906 0.43921906 0.12549116 0.3137279  0.43921906]

    Quantum resource estimation can be performed in the same way as in the first example:

    ::

        @count_ops(meas_behavior="0")
        def main():

            x = CKS(BA, 0.01, kappa).apply_rus(prep_b)()
            return x

        res_dict = main()
        print(res_dict)
        # {'gphase': 51, 's_dg': 70, 's': 120, 'z': 12, 'x': 127, 't': 298, 'u3': 157, 'cx': 1194, 'p': 118, 'ry': 2, 'h': 390, 't_dg': 348, 'measure': 116}

    """

    j_0, beta = cks_params(eps, kappa, max_beta=max_beta)
    cheb_coeffs = cks_coeffs(j_0, beta)

    m = len(A._anc_templates)

    # Following https://math.berkeley.edu/~linlin/qasc/qasc_notes.pdf (page 104):
    # T1 = (U R), T2 = (U R U_dg R) -> T2^k T1 is block encoding T_{2k+1}(A)
    # This is valid even if the block encoding unitary is not Hermitian.
    # The first control-(U R) can be simplified to U since:
    # 1) The first qubit of the |unary> state is always 1.
    # 2) The R acts as identity on |0>.

    def inv_unitary(*args):
        with invert():
            A.unitary(*args)

    def T2(*args):
        reflection(args[:m])
        with conjugate(inv_unitary)(*args):
            reflection(args[:m])

    def new_unitary(*args):
        # Core LCU protocol: PREP, SELECT, PREP^†
        with conjugate(unary_prep)(args[0], cheb_coeffs):

            A.unitary(*args[1:])

            for i in jrange(1, j_0 + 1):
                z(args[0][i])
                with control(args[0][i]):
                    T2(*args[1:])

    new_anc_templates = [QuantumFloat(j_0 + 1).template()] + A._anc_templates
    new_alpha = np.sum(np.abs(cheb_coeffs)) / A.alpha
    return BlockEncoding(new_alpha, new_anc_templates, new_unitary)



# Autogenerated by mlir-tblgen; don't manually edit.

from ._ods_common import _cext as _ods_cext
from ._ods_common import (
    equally_sized_accessor as _ods_equally_sized_accessor,
    get_default_loc_context as _ods_get_default_loc_context,
    get_op_results_or_values as _get_op_results_or_values,
    segmented_accessor as _ods_segmented_accessor,
)

_ods_ir = _ods_cext.ir

import builtins
from typing import Sequence as _Sequence, Union as _Union, Optional as _Optional


@_ods_cext.register_dialect
class _Dialect(_ods_ir.Dialect):
    DIALECT_NAMESPACE = "jasp"


@_ods_cext.register_operation(_Dialect)
class ConsumeQuantumKernelOp(_ods_ir.OpView):
    r"""
    Indicates to the execution environment that the quantum computation has concluded.
    """

    OPERATION_NAME = "jasp.consume_quantum_kernel"

    _ODS_REGIONS = (0, True)

    def __init__(self, success, qst, *, loc=None, ip=None):
        operands = []
        attributes = {}
        regions = None
        operands.append(qst)
        _ods_context = _ods_get_default_loc_context(loc)
        results = []
        results.append(success)
        _ods_successors = None
        super().__init__(
            self.OPERATION_NAME,
            self._ODS_REGIONS,
            self._ODS_OPERAND_SEGMENTS,
            self._ODS_RESULT_SEGMENTS,
            attributes=attributes,
            results=results,
            operands=operands,
            successors=_ods_successors,
            regions=regions,
            loc=loc,
            ip=ip,
        )

    @builtins.property
    def qst(self) -> _ods_ir.Value:
        return self.operation.operands[0]

    @builtins.property
    def success(self) -> _ods_ir.OpResult:
        return self.operation.results[0]


def consume_quantum_kernel(success, qst, *, loc=None, ip=None) -> _ods_ir.OpResult:
    return ConsumeQuantumKernelOp(success=success, qst=qst, loc=loc, ip=ip).result


@_ods_cext.register_operation(_Dialect)
class CreateQuantumKernelOp(_ods_ir.OpView):
    r"""
    Indicates to the execution environment that a quantum computation will start.
    """

    OPERATION_NAME = "jasp.create_quantum_kernel"

    _ODS_REGIONS = (0, True)

    def __init__(self, result, *, loc=None, ip=None):
        operands = []
        attributes = {}
        regions = None
        _ods_context = _ods_get_default_loc_context(loc)
        results = []
        results.append(result)
        _ods_successors = None
        super().__init__(
            self.OPERATION_NAME,
            self._ODS_REGIONS,
            self._ODS_OPERAND_SEGMENTS,
            self._ODS_RESULT_SEGMENTS,
            attributes=attributes,
            results=results,
            operands=operands,
            successors=_ods_successors,
            regions=regions,
            loc=loc,
            ip=ip,
        )

    @builtins.property
    def result(self) -> _ods_ir.OpResult:
        return self.operation.results[0]


def create_quantum_kernel(result, *, loc=None, ip=None) -> _ods_ir.OpResult:
    return CreateQuantumKernelOp(result=result, loc=loc, ip=ip).result


@_ods_cext.register_operation(_Dialect)
class CreateQubitsOp(_ods_ir.OpView):
    r"""
    Allocates a QubitArray containing n qubits. N can be dynamically sized
    """

    OPERATION_NAME = "jasp.create_qubits"

    _ODS_REGIONS = (0, True)

    def __init__(self, result, qst_out, amount, qst_in, *, loc=None, ip=None):
        operands = []
        attributes = {}
        regions = None
        operands.append(amount)
        operands.append(qst_in)
        _ods_context = _ods_get_default_loc_context(loc)
        results = []
        results.append(result)
        results.append(qst_out)
        _ods_successors = None
        super().__init__(
            self.OPERATION_NAME,
            self._ODS_REGIONS,
            self._ODS_OPERAND_SEGMENTS,
            self._ODS_RESULT_SEGMENTS,
            attributes=attributes,
            results=results,
            operands=operands,
            successors=_ods_successors,
            regions=regions,
            loc=loc,
            ip=ip,
        )

    @builtins.property
    def amount(self) -> _ods_ir.Value:
        return self.operation.operands[0]

    @builtins.property
    def qst_in(self) -> _ods_ir.Value:
        return self.operation.operands[1]

    @builtins.property
    def result(self) -> _ods_ir.OpResult:
        return self.operation.results[0]

    @builtins.property
    def qst_out(self) -> _ods_ir.OpResult:
        return self.operation.results[1]


def create_qubits(
    result, qst_out, amount, qst_in, *, loc=None, ip=None
) -> _ods_ir.OpResultList:
    return CreateQubitsOp(
        result=result, qst_out=qst_out, amount=amount, qst_in=qst_in, loc=loc, ip=ip
    ).results


@_ods_cext.register_operation(_Dialect)
class DeleteQubitsOp(_ods_ir.OpView):
    r"""
    Indicates to the execution environment that the corresponding qubits can be reused.
    """

    OPERATION_NAME = "jasp.delete_qubits"

    _ODS_REGIONS = (0, True)

    def __init__(self, out_qst, qubits, in_qst, *, loc=None, ip=None):
        operands = []
        attributes = {}
        regions = None
        operands.append(qubits)
        operands.append(in_qst)
        _ods_context = _ods_get_default_loc_context(loc)
        results = []
        results.append(out_qst)
        _ods_successors = None
        super().__init__(
            self.OPERATION_NAME,
            self._ODS_REGIONS,
            self._ODS_OPERAND_SEGMENTS,
            self._ODS_RESULT_SEGMENTS,
            attributes=attributes,
            results=results,
            operands=operands,
            successors=_ods_successors,
            regions=regions,
            loc=loc,
            ip=ip,
        )

    @builtins.property
    def qubits(self) -> _ods_ir.Value:
        return self.operation.operands[0]

    @builtins.property
    def in_qst(self) -> _ods_ir.Value:
        return self.operation.operands[1]

    @builtins.property
    def out_qst(self) -> _ods_ir.OpResult:
        return self.operation.results[0]


def delete_qubits(out_qst, qubits, in_qst, *, loc=None, ip=None) -> _ods_ir.OpResult:
    return DeleteQubitsOp(
        out_qst=out_qst, qubits=qubits, in_qst=in_qst, loc=loc, ip=ip
    ).result


@_ods_cext.register_operation(_Dialect)
class FuseOp(_ods_ir.OpView):
    r"""
    Fuses two QubitArrays, Qubits, or combinations thereof to create a larger QubitArray.
    """

    OPERATION_NAME = "jasp.fuse"

    _ODS_REGIONS = (0, True)

    def __init__(self, result, operand1, operand2, *, loc=None, ip=None):
        operands = []
        attributes = {}
        regions = None
        operands.append(operand1)
        operands.append(operand2)
        _ods_context = _ods_get_default_loc_context(loc)
        results = []
        results.append(result)
        _ods_successors = None
        super().__init__(
            self.OPERATION_NAME,
            self._ODS_REGIONS,
            self._ODS_OPERAND_SEGMENTS,
            self._ODS_RESULT_SEGMENTS,
            attributes=attributes,
            results=results,
            operands=operands,
            successors=_ods_successors,
            regions=regions,
            loc=loc,
            ip=ip,
        )

    @builtins.property
    def operand1(self) -> _ods_ir.Value:
        return self.operation.operands[0]

    @builtins.property
    def operand2(self) -> _ods_ir.Value:
        return self.operation.operands[1]

    @builtins.property
    def result(self) -> _ods_ir.OpResult:
        return self.operation.results[0]


def fuse(result, operand1, operand2, *, loc=None, ip=None) -> _ods_ir.OpResult:
    return FuseOp(
        result=result, operand1=operand1, operand2=operand2, loc=loc, ip=ip
    ).result


@_ods_cext.register_operation(_Dialect)
class GetQubitOp(_ods_ir.OpView):
    r"""
    Retrieves a single qubit from a given QubitArray at position.
    """

    OPERATION_NAME = "jasp.get_qubit"

    _ODS_REGIONS = (0, True)

    def __init__(self, result, qb_array, position, *, loc=None, ip=None):
        operands = []
        attributes = {}
        regions = None
        operands.append(qb_array)
        operands.append(position)
        _ods_context = _ods_get_default_loc_context(loc)
        results = []
        results.append(result)
        _ods_successors = None
        super().__init__(
            self.OPERATION_NAME,
            self._ODS_REGIONS,
            self._ODS_OPERAND_SEGMENTS,
            self._ODS_RESULT_SEGMENTS,
            attributes=attributes,
            results=results,
            operands=operands,
            successors=_ods_successors,
            regions=regions,
            loc=loc,
            ip=ip,
        )

    @builtins.property
    def qb_array(self) -> _ods_ir.Value:
        return self.operation.operands[0]

    @builtins.property
    def position(self) -> _ods_ir.Value:
        return self.operation.operands[1]

    @builtins.property
    def result(self) -> _ods_ir.OpResult:
        return self.operation.results[0]


def get_qubit(result, qb_array, position, *, loc=None, ip=None) -> _ods_ir.OpResult:
    return GetQubitOp(
        result=result, qb_array=qb_array, position=position, loc=loc, ip=ip
    ).result


@_ods_cext.register_operation(_Dialect)
class GetSizeOp(_ods_ir.OpView):
    r"""
    Returns the number of qubits in a given QubitArray.
    """

    OPERATION_NAME = "jasp.get_size"

    _ODS_REGIONS = (0, True)

    def __init__(self, size, qb_array, *, loc=None, ip=None):
        operands = []
        attributes = {}
        regions = None
        operands.append(qb_array)
        _ods_context = _ods_get_default_loc_context(loc)
        results = []
        results.append(size)
        _ods_successors = None
        super().__init__(
            self.OPERATION_NAME,
            self._ODS_REGIONS,
            self._ODS_OPERAND_SEGMENTS,
            self._ODS_RESULT_SEGMENTS,
            attributes=attributes,
            results=results,
            operands=operands,
            successors=_ods_successors,
            regions=regions,
            loc=loc,
            ip=ip,
        )

    @builtins.property
    def qb_array(self) -> _ods_ir.Value:
        return self.operation.operands[0]

    @builtins.property
    def size(self) -> _ods_ir.OpResult:
        return self.operation.results[0]


def get_size(size, qb_array, *, loc=None, ip=None) -> _ods_ir.OpResult:
    return GetSizeOp(size=size, qb_array=qb_array, loc=loc, ip=ip).result


@_ods_cext.register_operation(_Dialect)
class MeasureOp(_ods_ir.OpView):
    r"""
    Performs a measurement of a given quantum state on a given qubit or qubit array.
    """

    OPERATION_NAME = "jasp.measure"

    _ODS_REGIONS = (0, True)

    def __init__(self, meas_res, out_qst, meas_q, in_qst, *, loc=None, ip=None):
        operands = []
        attributes = {}
        regions = None
        operands.append(meas_q)
        operands.append(in_qst)
        _ods_context = _ods_get_default_loc_context(loc)
        results = []
        results.append(meas_res)
        results.append(out_qst)
        _ods_successors = None
        super().__init__(
            self.OPERATION_NAME,
            self._ODS_REGIONS,
            self._ODS_OPERAND_SEGMENTS,
            self._ODS_RESULT_SEGMENTS,
            attributes=attributes,
            results=results,
            operands=operands,
            successors=_ods_successors,
            regions=regions,
            loc=loc,
            ip=ip,
        )

    @builtins.property
    def meas_q(self) -> _ods_ir.Value:
        return self.operation.operands[0]

    @builtins.property
    def in_qst(self) -> _ods_ir.Value:
        return self.operation.operands[1]

    @builtins.property
    def meas_res(self) -> _ods_ir.OpResult:
        return self.operation.results[0]

    @builtins.property
    def out_qst(self) -> _ods_ir.OpResult:
        return self.operation.results[1]


def measure(
    meas_res, out_qst, meas_q, in_qst, *, loc=None, ip=None
) -> _ods_ir.OpResultList:
    return MeasureOp(
        meas_res=meas_res, out_qst=out_qst, meas_q=meas_q, in_qst=in_qst, loc=loc, ip=ip
    ).results


@_ods_cext.register_operation(_Dialect)
class ParityOp(_ods_ir.OpView):
    r"""
    Computes the parity (XOR sum) of a set of measurement results. Supports expectation and observable attributes for error correction contexts.
    """

    OPERATION_NAME = "jasp.parity"

    _ODS_REGIONS = (0, True)

    def __init__(
        self, result, measurements, expectation, observable, *, loc=None, ip=None
    ):
        operands = []
        attributes = {}
        regions = None
        operands.extend(_get_op_results_or_values(measurements))
        _ods_context = _ods_get_default_loc_context(loc)
        attributes["expectation"] = (
            expectation
            if (
                isinstance(expectation, _ods_ir.Attribute)
                or not _ods_ir.AttrBuilder.contains("I64Attr")
            )
            else _ods_ir.AttrBuilder.get("I64Attr")(expectation, context=_ods_context)
        )
        attributes["observable"] = (
            observable
            if (
                isinstance(observable, _ods_ir.Attribute)
                or not _ods_ir.AttrBuilder.contains("I64Attr")
            )
            else _ods_ir.AttrBuilder.get("I64Attr")(observable, context=_ods_context)
        )
        results = []
        results.append(result)
        _ods_successors = None
        super().__init__(
            self.OPERATION_NAME,
            self._ODS_REGIONS,
            self._ODS_OPERAND_SEGMENTS,
            self._ODS_RESULT_SEGMENTS,
            attributes=attributes,
            results=results,
            operands=operands,
            successors=_ods_successors,
            regions=regions,
            loc=loc,
            ip=ip,
        )

    @builtins.property
    def measurements(self) -> _ods_ir.OpOperandList:
        _ods_variadic_group_length = len(self.operation.operands) - 1 + 1
        return self.operation.operands[0 : 0 + _ods_variadic_group_length]

    @builtins.property
    def expectation(self) -> _ods_ir.IntegerAttr:
        return self.operation.attributes["expectation"]

    @expectation.setter
    def expectation(self, value: _ods_ir.IntegerAttr):
        if value is None:
            raise ValueError("'None' not allowed as value for mandatory attributes")
        self.operation.attributes["expectation"] = value

    @builtins.property
    def observable(self) -> _ods_ir.IntegerAttr:
        return self.operation.attributes["observable"]

    @observable.setter
    def observable(self, value: _ods_ir.IntegerAttr):
        if value is None:
            raise ValueError("'None' not allowed as value for mandatory attributes")
        self.operation.attributes["observable"] = value

    @builtins.property
    def result(self) -> _ods_ir.OpResult:
        return self.operation.results[0]


def parity(
    result, measurements, expectation, observable, *, loc=None, ip=None
) -> _ods_ir.OpResult:
    return ParityOp(
        result=result,
        measurements=measurements,
        expectation=expectation,
        observable=observable,
        loc=loc,
        ip=ip,
    ).result


@_ods_cext.register_operation(_Dialect)
class QuantumGateOp(_ods_ir.OpView):
    r"""
    This operation enables quantum processing of quantum states with (parametric) gates
    """

    OPERATION_NAME = "jasp.quantum_gate"

    _ODS_REGIONS = (0, True)

    def __init__(self, out_qst, gate_type, gate_operands, in_qst, *, loc=None, ip=None):
        operands = []
        attributes = {}
        regions = None
        operands.extend(_get_op_results_or_values(gate_operands))
        operands.append(in_qst)
        _ods_context = _ods_get_default_loc_context(loc)
        attributes["gate_type"] = (
            gate_type
            if (
                isinstance(gate_type, _ods_ir.Attribute)
                or not _ods_ir.AttrBuilder.contains("StrAttr")
            )
            else _ods_ir.AttrBuilder.get("StrAttr")(gate_type, context=_ods_context)
        )
        results = []
        results.append(out_qst)
        _ods_successors = None
        super().__init__(
            self.OPERATION_NAME,
            self._ODS_REGIONS,
            self._ODS_OPERAND_SEGMENTS,
            self._ODS_RESULT_SEGMENTS,
            attributes=attributes,
            results=results,
            operands=operands,
            successors=_ods_successors,
            regions=regions,
            loc=loc,
            ip=ip,
        )

    @builtins.property
    def gate_operands(self) -> _ods_ir.OpOperandList:
        _ods_variadic_group_length = len(self.operation.operands) - 2 + 1
        return self.operation.operands[0 : 0 + _ods_variadic_group_length]

    @builtins.property
    def in_qst(self) -> _ods_ir.Value:
        _ods_variadic_group_length = len(self.operation.operands) - 2 + 1
        return self.operation.operands[1 + _ods_variadic_group_length - 1]

    @builtins.property
    def gate_type(self) -> _ods_ir.StringAttr:
        return self.operation.attributes["gate_type"]

    @gate_type.setter
    def gate_type(self, value: _ods_ir.StringAttr):
        if value is None:
            raise ValueError("'None' not allowed as value for mandatory attributes")
        self.operation.attributes["gate_type"] = value

    @builtins.property
    def out_qst(self) -> _ods_ir.OpResult:
        return self.operation.results[0]


def quantum_gate(
    out_qst, gate_type, gate_operands, in_qst, *, loc=None, ip=None
) -> _ods_ir.OpResult:
    return QuantumGateOp(
        out_qst=out_qst,
        gate_type=gate_type,
        gate_operands=gate_operands,
        in_qst=in_qst,
        loc=loc,
        ip=ip,
    ).result


@_ods_cext.register_operation(_Dialect)
class ResetOp(_ods_ir.OpView):
    r"""
    Performs a reset operation on a single qubit or qubit array, returning them to the |0⟩ state.
    """

    OPERATION_NAME = "jasp.reset"

    _ODS_REGIONS = (0, True)

    def __init__(self, out_qst, qubits, in_qst, *, loc=None, ip=None):
        operands = []
        attributes = {}
        regions = None
        operands.append(qubits)
        operands.append(in_qst)
        _ods_context = _ods_get_default_loc_context(loc)
        results = []
        results.append(out_qst)
        _ods_successors = None
        super().__init__(
            self.OPERATION_NAME,
            self._ODS_REGIONS,
            self._ODS_OPERAND_SEGMENTS,
            self._ODS_RESULT_SEGMENTS,
            attributes=attributes,
            results=results,
            operands=operands,
            successors=_ods_successors,
            regions=regions,
            loc=loc,
            ip=ip,
        )

    @builtins.property
    def qubits(self) -> _ods_ir.Value:
        return self.operation.operands[0]

    @builtins.property
    def in_qst(self) -> _ods_ir.Value:
        return self.operation.operands[1]

    @builtins.property
    def out_qst(self) -> _ods_ir.OpResult:
        return self.operation.results[0]


def reset(out_qst, qubits, in_qst, *, loc=None, ip=None) -> _ods_ir.OpResult:
    return ResetOp(out_qst=out_qst, qubits=qubits, in_qst=in_qst, loc=loc, ip=ip).result


@_ods_cext.register_operation(_Dialect)
class SliceOp(_ods_ir.OpView):
    r"""
    Returns a subset of qubits from a QubitArray using start and end indices.
    """

    OPERATION_NAME = "jasp.slice"

    _ODS_REGIONS = (0, True)

    def __init__(self, result, qb_array, start, end, *, loc=None, ip=None):
        operands = []
        attributes = {}
        regions = None
        operands.append(qb_array)
        operands.append(start)
        operands.append(end)
        _ods_context = _ods_get_default_loc_context(loc)
        results = []
        results.append(result)
        _ods_successors = None
        super().__init__(
            self.OPERATION_NAME,
            self._ODS_REGIONS,
            self._ODS_OPERAND_SEGMENTS,
            self._ODS_RESULT_SEGMENTS,
            attributes=attributes,
            results=results,
            operands=operands,
            successors=_ods_successors,
            regions=regions,
            loc=loc,
            ip=ip,
        )

    @builtins.property
    def qb_array(self) -> _ods_ir.Value:
        return self.operation.operands[0]

    @builtins.property
    def start(self) -> _ods_ir.Value:
        return self.operation.operands[1]

    @builtins.property
    def end(self) -> _ods_ir.Value:
        return self.operation.operands[2]

    @builtins.property
    def result(self) -> _ods_ir.OpResult:
        return self.operation.results[0]


def slice(result, qb_array, start, end, *, loc=None, ip=None) -> _ods_ir.OpResult:
    return SliceOp(
        result=result, qb_array=qb_array, start=start, end=end, loc=loc, ip=ip
    ).result
